from bloqade import stim


def test_wrapper_x():

    @stim.main
    def main_x():
        stim.X(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_x_wrap():
        stim.x(targets=(0, 1, 2), dagger=False)

    assert main_x.callable_region.is_structurally_equal(main_x_wrap.callable_region)


def test_wrapper_y():

    @stim.main
    def main_y():
        stim.Y(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_y_wrap():
        stim.y(targets=(0, 1, 2), dagger=False)

    assert main_y.callable_region.is_structurally_equal(main_y_wrap.callable_region)


def test_wrapper_z():

    @stim.main
    def main_z():
        stim.Z(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_z_wrap():
        stim.z(targets=(0, 1, 2), dagger=False)

    assert main_z.callable_region.is_structurally_equal(main_z_wrap.callable_region)


def test_wrapper_identity():

    @stim.main
    def main_identity():
        stim.Identity(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_identity_wrap():
        stim.identity(targets=(0, 1, 2), dagger=False)

    assert main_identity.callable_region.is_structurally_equal(
        main_identity_wrap.callable_region
    )


def test_wrapper_h():

    @stim.main
    def main_h():
        stim.H(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_h_wrap():
        stim.h(targets=(0, 1, 2), dagger=False)

    assert main_h.callable_region.is_structurally_equal(main_h_wrap.callable_region)


def test_wrapper_s():

    @stim.main
    def main_s():
        stim.S(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_s_wrap():
        stim.s(targets=(0, 1, 2), dagger=False)

    assert main_s.callable_region.is_structurally_equal(main_s_wrap.callable_region)


def test_wrapper_sqrt_x():

    @stim.main
    def main_sqrt_x():
        stim.SqrtX(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_sqrt_x_wrap():
        stim.sqrt_x(targets=(0, 1, 2), dagger=False)

    assert main_sqrt_x.callable_region.is_structurally_equal(
        main_sqrt_x_wrap.callable_region
    )


def test_wrapper_sqrt_y():

    @stim.main
    def main_sqrt_y():
        stim.SqrtY(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_sqrt_y_wrap():
        stim.sqrt_y(targets=(0, 1, 2), dagger=False)

    assert main_sqrt_y.callable_region.is_structurally_equal(
        main_sqrt_y_wrap.callable_region
    )


def test_wrapper_sqrt_z():

    @stim.main
    def main_sqrt_z():
        stim.SqrtZ(targets=(0, 1, 2), dagger=False)

    @stim.main
    def main_sqrt_z_wrap():
        stim.sqrt_z(targets=(0, 1, 2), dagger=False)

    assert main_sqrt_z.callable_region.is_structurally_equal(
        main_sqrt_z_wrap.callable_region
    )


def test_wrapper_swap():

    @stim.main
    def main_swap():
        stim.Swap(targets=(0, 3, 1, 4, 2, 5), dagger=False)

    @stim.main
    def main_swap_wrap():
        stim.swap(targets=(0, 3, 1, 4, 2, 5), dagger=False)

    assert main_swap.callable_region.is_structurally_equal(
        main_swap_wrap.callable_region
    )


def test_wrapper_cx():

    @stim.main
    def main_cx():
        stim.CX(controls=(0, 1, 2), targets=(3, 4, 5), dagger=False)

    @stim.main
    def main_cx_wrap():
        stim.cx(controls=(0, 1, 2), targets=(3, 4, 5), dagger=False)

    assert main_cx.callable_region.is_structurally_equal(main_cx_wrap.callable_region)


def test_wrapper_cy():

    @stim.main
    def main_cy():
        stim.CY(controls=(0, 1, 2), targets=(3, 4, 5), dagger=False)

    @stim.main
    def main_cy_wrap():
        stim.cy(controls=(0, 1, 2), targets=(3, 4, 5), dagger=False)

    assert main_cy.callable_region.is_structurally_equal(main_cy_wrap.callable_region)


def test_wrapper_cz():

    @stim.main
    def main_cz():
        stim.CZ(controls=(0, 1, 2), targets=(3, 4, 5), dagger=False)

    @stim.main
    def main_cz_wrap():
        stim.cz(controls=(0, 1, 2), targets=(3, 4, 5), dagger=False)

    assert main_cz.callable_region.is_structurally_equal(main_cz_wrap.callable_region)


def test_wrap_pauli_string():

    @stim.main
    def main_pauli_string():
        stim.NewPauliString(
            string=("X", "Z", "Y"), flipped=(False, False, False), targets=(0, 1, 2)
        )

    @stim.main
    def main_pauli_string_wrap():
        stim.pauli_string(
            string=("X", "Z", "Y"), flipped=(False, False, False), targets=(0, 1, 2)
        )

    assert main_pauli_string.callable_region.is_structurally_equal(
        main_pauli_string_wrap.callable_region
    )


def test_wrap_spp():

    @stim.main
    def main_spp():
        stim.SPP(
            targets=(
                stim.pauli_string(
                    string=("X", "Z", "Y"),
                    flipped=(False, False, False),
                    targets=(0, 1, 2),
                ),
                stim.pauli_string(
                    string=("Z", "Z", "Z"),
                    flipped=(False, False, False),
                    targets=(3, 4, 5),
                ),
            ),
            dagger=False,
        )

    @stim.main
    def main_spp_wrap():
        stim.spp(
            targets=(
                stim.pauli_string(
                    string=("X", "Z", "Y"),
                    flipped=(False, False, False),
                    targets=(0, 1, 2),
                ),
                stim.pauli_string(
                    string=("Z", "Z", "Z"),
                    flipped=(False, False, False),
                    targets=(3, 4, 5),
                ),
            ),
            dagger=False,
        )

    assert main_spp.callable_region.is_structurally_equal(main_spp_wrap.callable_region)


def test_wrap_rec():

    @stim.main
    def main_rec():
        stim.GetRecord(0)

    @stim.main
    def main_rec_wrap():
        stim.rec(0)

    assert main_rec.callable_region.is_structurally_equal(main_rec_wrap.callable_region)


def test_wrap_detector():

    @stim.main
    def main_detector():
        stim.Detector(
            coord=(0.0, 1.0, 2.0), targets=(stim.rec(0), stim.rec(1), stim.rec(2))
        )

    @stim.main
    def main_detector_wrap():
        stim.detector(
            coord=(0.0, 1.0, 2.0), targets=(stim.rec(0), stim.rec(1), stim.rec(2))
        )

    assert main_detector.callable_region.is_structurally_equal(
        main_detector_wrap.callable_region
    )


def test_wrap_observable_include():

    @stim.main
    def main_observable_include():
        stim.ObservableInclude(idx=0, targets=(stim.rec(0), stim.rec(1), stim.rec(2)))

    @stim.main
    def main_observable_include_wrap():
        stim.observable_include(idx=0, targets=(stim.rec(0), stim.rec(1), stim.rec(2)))

    assert main_observable_include.callable_region.is_structurally_equal(
        main_observable_include_wrap.callable_region
    )


def test_wrap_tick():

    @stim.main
    def main_tick():
        stim.Tick()

    @stim.main
    def main_tick_wrap():
        stim.tick()

    assert main_tick.callable_region.is_structurally_equal(
        main_tick_wrap.callable_region
    )


def test_wrap_mz():

    @stim.main
    def main_mz():
        stim.MZ(p=0.5, targets=(0, 1, 2))

    @stim.main
    def main_mz_wrap():
        stim.mz(p=0.5, targets=(0, 1, 2))

    assert main_mz.callable_region.is_structurally_equal(main_mz_wrap.callable_region)


def test_wrap_my():

    @stim.main
    def main_my():
        stim.MY(p=0.5, targets=(0, 1, 2))

    @stim.main
    def main_my_wrap():
        stim.my(p=0.5, targets=(0, 1, 2))

    assert main_my.callable_region.is_structurally_equal(main_my_wrap.callable_region)


def test_wrap_mx():

    @stim.main
    def main_mx():
        stim.MX(p=0.5, targets=(0, 1, 2))

    @stim.main
    def main_mx_wrap():
        stim.mx(p=0.5, targets=(0, 1, 2))

    assert main_mx.callable_region.is_structurally_equal(main_mx_wrap.callable_region)


def test_wrap_mzz():

    @stim.main
    def main_mzz():
        stim.MZZ(p=0.5, targets=(0, 1, 2))

    @stim.main
    def main_mzz_wrap():
        stim.mzz(p=0.5, targets=(0, 1, 2))

    assert main_mzz.callable_region.is_structurally_equal(main_mzz_wrap.callable_region)


def test_wrap_myy():

    @stim.main
    def main_myy():
        stim.MYY(p=0.5, targets=(0, 1, 2))

    @stim.main
    def main_myy_wrap():
        stim.myy(p=0.5, targets=(0, 1, 2))

    assert main_myy.callable_region.is_structurally_equal(main_myy_wrap.callable_region)


def test_wrap_mxx():

    @stim.main
    def main_mxx():
        stim.MXX(p=0.5, targets=(0, 1, 2))

    @stim.main
    def main_mxx_wrap():
        stim.mxx(p=0.5, targets=(0, 1, 2))

    assert main_mxx.callable_region.is_structurally_equal(main_mxx_wrap.callable_region)


def test_wrap_mpp():

    @stim.main
    def main_mpp():
        stim.PPMeasurement(
            p=0.5,
            targets=(
                stim.pauli_string(
                    string=("X", "Z", "Y"),
                    flipped=(False, False, False),
                    targets=(0, 1, 2),
                ),
                stim.pauli_string(
                    string=("Z", "Z", "Z"),
                    flipped=(False, False, False),
                    targets=(3, 4, 5),
                ),
            ),
        )

    @stim.main
    def main_mpp_wrap():
        stim.mpp(
            p=0.5,
            targets=(
                stim.pauli_string(
                    string=("X", "Z", "Y"),
                    flipped=(False, False, False),
                    targets=(0, 1, 2),
                ),
                stim.pauli_string(
                    string=("Z", "Z", "Z"),
                    flipped=(False, False, False),
                    targets=(3, 4, 5),
                ),
            ),
        )

    assert main_mpp.callable_region.is_structurally_equal(main_mpp_wrap.callable_region)


def test_wrap_rz():

    @stim.main
    def main_rz():
        stim.RZ(targets=(0, 1, 2))

    @stim.main
    def main_rz_wrap():
        stim.rz(targets=(0, 1, 2))

    assert main_rz.callable_region.is_structurally_equal(main_rz_wrap.callable_region)


def test_wrap_rx():

    @stim.main
    def main_rx():
        stim.RX(targets=(0, 1, 2))

    @stim.main
    def main_rx_wrap():
        stim.rx(targets=(0, 1, 2))

    assert main_rx.callable_region.is_structurally_equal(main_rx_wrap.callable_region)


def test_wrap_ry():

    @stim.main
    def main_ry():
        stim.RY(targets=(0, 1, 2))

    @stim.main
    def main_ry_wrap():
        stim.ry(targets=(0, 1, 2))

    assert main_ry.callable_region.is_structurally_equal(main_ry_wrap.callable_region)



from kirin import ir, types
from kirin.decl import info, statement
from bloqade.qasm2.types import QubitType

dialect = ir.Dialect("qasm2.noise")


@statement(dialect=dialect)
class Pauli1(ir.Statement):
    name = "pauli_1"
    traits = frozenset({ir.FromPythonCall()})
    px: ir.SSAValue = info.argument(types.Float)
    py: ir.SSAValue = info.argument(types.Float)
    pz: ir.SSAValue = info.argument(types.Float)
    qarg: ir.SSAValue = info.argument(QubitType)



"""
qubit.address method table for a few builtin dialects.
"""

from kirin import interp
from kirin.analysis import ForwardFrame, const
from kirin.dialects import cf, py, scf, func, ilist

from .lattice import Address, NotQubit, AddressReg, AddressQubit, AddressTuple
from .analysis import AddressAnalysis


@py.binop.dialect.register(key="qubit.address")
class PyBinOp(interp.MethodTable):

    @interp.impl(py.Add)
    def add(self, interp: AddressAnalysis, frame: interp.Frame, stmt: py.Add):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)

        if isinstance(lhs, AddressTuple) and isinstance(rhs, AddressTuple):
            return (AddressTuple(data=lhs.data + rhs.data),)
        else:
            return (NotQubit(),)


@py.tuple.dialect.register(key="qubit.address")
class PyTuple(interp.MethodTable):
    @interp.impl(py.tuple.New)
    def new_tuple(
        self,
        interp: AddressAnalysis,
        frame: interp.Frame,
        stmt: py.tuple.New,
    ):
        return (AddressTuple(frame.get_values(stmt.args)),)


@ilist.dialect.register(key="qubit.address")
class IList(interp.MethodTable):
    @interp.impl(ilist.New)
    def new_ilist(
        self,
        interp: AddressAnalysis,
        frame: interp.Frame,
        stmt: ilist.New,
    ):
        return (AddressTuple(frame.get_values(stmt.args)),)


@py.list.dialect.register(key="qubit.address")
class PyList(interp.MethodTable):
    @interp.impl(py.list.New)
    def new_ilist(
        self,
        interp: AddressAnalysis,
        frame: interp.Frame,
        stmt: py.list.New,
    ):
        return (AddressTuple(frame.get_values(stmt.args)),)


@py.indexing.dialect.register(key="qubit.address")
class PyIndexing(interp.MethodTable):
    @interp.impl(py.GetItem)
    def getitem(self, interp: AddressAnalysis, frame: interp.Frame, stmt: py.GetItem):
        idx = interp.get_const_value(int, stmt.index)
        obj = frame.get(stmt.obj)
        if isinstance(obj, AddressTuple):
            return (obj.data[idx],)
        elif isinstance(obj, AddressReg):
            return (AddressQubit(obj.data[idx]),)
        else:
            return (NotQubit(),)


@py.assign.dialect.register(key="qubit.address")
class PyAssign(interp.MethodTable):
    @interp.impl(py.Alias)
    def alias(self, interp: AddressAnalysis, frame: interp.Frame, stmt: py.Alias):
        return (frame.get(stmt.value),)


@func.dialect.register(key="qubit.address")
class Func(interp.MethodTable):
    @interp.impl(func.Return)
    def return_(self, _: AddressAnalysis, frame: interp.Frame, stmt: func.Return):
        return interp.ReturnValue(frame.get(stmt.value))

    # TODO: replace with the generic implementation
    @interp.impl(func.Invoke)
    def invoke(self, interp_: AddressAnalysis, frame: interp.Frame, stmt: func.Invoke):
        _, ret = interp_.run_method(
            stmt.callee,
            interp_.permute_values(
                stmt.callee.arg_names, frame.get_values(stmt.inputs), stmt.kwargs
            ),
        )
        return (ret,)

    # TODO: support lambda?


@cf.dialect.register(key="qubit.address")
class Cf(cf.typeinfer.TypeInfer):
    # NOTE: cf just re-use the type infer method table
    # it's the same process as type infer.
    pass


@scf.dialect.register(key="qubit.address")
class Scf(scf.absint.Methods):

    @interp.impl(scf.For)
    def for_loop(
        self,
        interp_: AddressAnalysis,
        frame: ForwardFrame[Address],
        stmt: scf.For,
    ):
        if not isinstance(hint := stmt.iterable.hints.get("const"), const.Value):
            return interp_.eval_stmt_fallback(frame, stmt)

        iterable = hint.data
        loop_vars = frame.get_values(stmt.initializers)
        body_block = stmt.body.blocks[0]
        block_args = body_block.args

        # NOTE: we need to actually run iteration in case there are
        # new allocations/re-assign in the loop body.
        for _ in iterable:
            with interp_.state.new_frame(interp_.new_frame(stmt)) as body_frame:
                body_frame.entries.update(frame.entries)
                body_frame.set_values(
                    block_args,
                    (NotQubit(),) + loop_vars,
                )
                loop_vars = interp_.run_ssacfg_region(body_frame, stmt.body)

            if loop_vars is None:
                loop_vars = ()
            elif isinstance(loop_vars, interp.ReturnValue):
                return loop_vars

        if isinstance(body_block.last_stmt, func.Return):
            frame.worklist.append(interp.Successor(body_block, NotQubit(), *loop_vars))
            return  # if terminate is Return, there is no result

        return loop_vars



from bloqade import stim

from .base import codegen


def test_cx():

    @stim.main
    def test_simple_cx():
        stim.CX(controls=(4, 5, 6, 7), targets=(0, 1, 2, 3), dagger=False)

    out = codegen(test_simple_cx)
    assert out.strip() == "CX 4 0 5 1 6 2 7 3"


def test_cx_cond_on_measure():

    @stim.main
    def test_simple_cx_cond_measure():
        stim.CX(
            controls=(stim.GetRecord(id=-1), 4, stim.GetRecord(id=-2)),
            targets=(0, 1, 2),
            dagger=False,
        )

    out = codegen(test_simple_cx_cond_measure)

    assert out.strip() == "CX rec[-1] 0 4 1 rec[-2] 2"



from typing import Optional

from bloqade import qasm2
from kirin.dialects import ilist
from bloqade.qbraid.target import qBraid


def test_qBraid_emit():

    @qasm2.main.add(qasm2.dialects.parallel).add(ilist)
    def main():
        qreg = qasm2.qreg(4)
        qasm2.cx(qreg[0], qreg[1])
        qasm2.reset(qreg[0])
        qasm2.parallel.cz(ctrls=[qreg[0], qreg[1]], qargs=[qreg[2], qreg[3]])

    class MockQBraidJob:

        def __init__(
            self, qasm: str, shots: Optional[int], tags: Optional[dict[str, str]]
        ):
            assert "KIRIN" in qasm
            assert "func" in qasm
            assert "lowering.call" in qasm
            assert "lowering.func" in qasm
            assert "py.ilist" in qasm
            assert "qasm2.core" in qasm
            assert "qasm2.expr" in qasm
            assert "qasm2.indexing" in qasm
            assert (
                'include "qelib1.inc";\nqreg qreg[4];\nCX qreg[0], qreg[1];\nreset qreg[0];\nparallel.CZ {\n  qreg[0], qreg[2];\n  qreg[1], qreg[3];\n}\n'
                in qasm
            )
            assert shots is None
            assert tags is None

    class MockDevice:

        def run(self, qasm: str, shots: Optional[int], tags: Optional[dict[str, str]]):
            return MockQBraidJob(qasm, shots, tags)

    class MockQBraidProvider:

        def get_device(self, api_key: str):
            assert api_key == "quera_qasm_simulator"

            return MockDevice()

    mock_provider = MockQBraidProvider()
    mock_qBraid_emitter = qBraid(
        main_target=qasm2.main.add(qasm2.dialects.parallel).add(ilist),
        gate_target=qasm2.gate.add(qasm2.dialects.parallel).add(ilist),
        provider=mock_provider,  # type: ignore
    )
    mock_qBraid_job = mock_qBraid_emitter.emit(method=main)

    assert isinstance(mock_qBraid_job, MockQBraidJob)



from typing import Any, Set, Dict, Iterable, Optional, final
from itertools import chain
from collections import OrderedDict
from dataclasses import field, dataclass
from collections.abc import Sequence

from kirin import ir, graph, interp, idtable
from kirin.lattice import (
    SingletonMeta,
    BoundedLattice,
    SimpleJoinMixin,
    SimpleMeetMixin,
)
from kirin.analysis import Forward, ForwardFrame
from kirin.dialects import func
from bloqade.analysis import address
from bloqade.qasm2.parse.print import Printer


@dataclass
class GateSchedule(
    SimpleJoinMixin["GateSchedule"],
    SimpleMeetMixin["GateSchedule"],
    BoundedLattice["GateSchedule"],
):

    @classmethod
    def bottom(cls) -> "GateSchedule":
        return NotQubit()

    @classmethod
    def top(cls) -> "GateSchedule":
        return Qubit()


@final
@dataclass
class NotQubit(GateSchedule, metaclass=SingletonMeta):

    def is_subseteq(self, other: GateSchedule) -> bool:
        return True


@final
@dataclass
class Qubit(GateSchedule, metaclass=SingletonMeta):

    def is_subseteq(self, other: GateSchedule) -> bool:
        return isinstance(other, Qubit)


# Treat global gates as terminators for this analysis, e.g. split block in half.


@dataclass(slots=True)
class StmtDag(graph.Graph[ir.Statement]):
    id_table: idtable.IdTable[ir.Statement] = field(
        default_factory=lambda: idtable.IdTable()
    )
    stmts: Dict[str, ir.Statement] = field(default_factory=OrderedDict)
    out_edges: Dict[str, Set[str]] = field(default_factory=OrderedDict)
    inc_edges: Dict[str, Set[str]] = field(default_factory=OrderedDict)

    def add_node(self, node: ir.Statement):
        node_id = self.id_table[node]
        self.stmts[node_id] = node
        self.out_edges.setdefault(node_id, set())
        self.inc_edges.setdefault(node_id, set())
        return node_id

    def add_edge(self, src: ir.Statement, dst: ir.Statement):
        src_id = self.add_node(src)
        dst_id = self.add_node(dst)

        self.out_edges[src_id].add(dst_id)
        self.inc_edges[dst_id].add(src_id)

    def get_parents(self, node: ir.Statement) -> Iterable[ir.Statement]:
        return (
            self.stmts[node_id]
            for node_id in self.inc_edges.get(self.id_table[node], set())
        )

    def get_children(self, node: ir.Statement) -> Iterable[ir.Statement]:
        return (
            self.stmts[node_id]
            for node_id in self.out_edges.get(self.id_table[node], set())
        )

    def get_neighbors(self, node: ir.Statement) -> Iterable[ir.Statement]:
        return chain(self.get_parents(node), self.get_children(node))

    def get_nodes(self) -> Iterable[ir.Statement]:
        return self.stmts.values()

    def get_edges(self) -> Iterable[tuple[ir.Statement, ir.Statement]]:
        return (
            (self.stmts[src], self.stmts[dst])
            for src, dsts in self.out_edges.items()
            for dst in dsts
        )

    def print(
        self,
        printer: Optional["Printer"] = None,
        analysis: dict["ir.SSAValue", Any] | None = None,
    ) -> None:
        raise NotImplementedError

    def topological_groups(self):
        """Split the dag into topological groups where each group
        contains nodes that have no dependencies on each other, but
        have dependencies on nodes in one or more previous groups.

        Yields:
            List[str]: A list of node ids in a topological group


        Raises:
            ValueError: If a cyclic dependency is detected


        The idea is to yield all nodes with no dependencies, then remove
        those nodes from the graph repeating until no nodes are left
        or we reach some upper limit. Worse case is a linear dag,
        so we can use len(dag.stmts) as the upper limit

        If we reach the limit and there are still nodes left, then we
        have a cyclic dependency.
        """

        inc_edges = {k: set(v) for k, v in self.inc_edges.items()}

        check_next = inc_edges.keys()

        for _ in range(len(self.stmts)):
            if len(inc_edges) == 0:
                break

            group = [node_id for node_id in check_next if len(inc_edges[node_id]) == 0]
            yield group

            check_next = set()
            for n in group:
                inc_edges.pop(n)
                for m in self.out_edges[n]:
                    check_next.add(m)
                    inc_edges[m].remove(n)

        if inc_edges:
            raise ValueError("Cyclic dependency detected")


@dataclass
class DagScheduleAnalysis(Forward[GateSchedule]):
    keys = ["qasm2.schedule.dag"]
    lattice = GateSchedule

    address_analysis: Dict[ir.SSAValue, address.Address]
    use_def: Dict[int, ir.Statement] = field(init=False)
    stmt_dag: StmtDag = field(init=False)
    stmt_dags: Dict[ir.Block, StmtDag] = field(init=False)

    def initialize(self):
        self.use_def = {}
        self.stmt_dag = StmtDag()
        self.stmt_dags = {}
        return super().initialize()

    def push_current_dag(self, block: ir.Block):
        # run when hitting terminator statements
        assert block not in self.stmt_dags, "Block already in stmt_dags"

        for node in self.use_def.values():
            self.stmt_dag.add_node(node)

        self.stmt_dags[block] = self.stmt_dag
        self.stmt_dag = StmtDag()
        self.use_def = {}

    def run_method(self, method: ir.Method, args: tuple[GateSchedule, ...]):
        # NOTE: we do not support dynamic calls here, thus no need to propagate method object
        return self.run_callable(method.code, (self.lattice.bottom(),) + args)

    def eval_stmt_fallback(self, frame: ForwardFrame, stmt: ir.Statement):
        if stmt.has_trait(ir.IsTerminator):
            assert (
                stmt.parent_block is not None
            ), "Terminator statement has no parent block"
            self.push_current_dag(stmt.parent_block)

        return tuple(self.lattice.top() for _ in stmt.results)

    def _update_dag(self, stmt: ir.Statement, addr: address.Address):
        if isinstance(addr, address.AddressQubit):
            old_stmt = self.use_def.get(addr.data, None)
            if old_stmt is not None:
                self.stmt_dag.add_edge(old_stmt, stmt)
            self.use_def[addr.data] = stmt
        elif isinstance(addr, address.AddressReg):
            for idx in addr.data:
                old_stmt = self.use_def.get(idx, None)
                if old_stmt is not None:
                    self.stmt_dag.add_edge(old_stmt, stmt)
                self.use_def[idx] = stmt
        elif isinstance(addr, address.AddressTuple):
            for sub_addr in addr.data:
                self._update_dag(stmt, sub_addr)

    def update_dag(self, stmt: ir.Statement, args: Sequence[ir.SSAValue]):
        for arg in args:
            self._update_dag(
                stmt, self.address_analysis.get(arg, address.Address.bottom())
            )

    def get_dags(self, mt: ir.Method, args=None, kwargs=None):
        if args is None:
            args = tuple(self.lattice.top() for _ in mt.args)

        self.run(mt, args, kwargs).expect()
        return self.stmt_dags


@func.dialect.register(key="qasm2.schedule.dag")
class FuncImpl(interp.MethodTable):
    @interp.impl(func.Invoke)
    @interp.impl(func.Call)
    def invoke(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: func.Invoke | func.Call,
    ):
        interp.update_dag(stmt, stmt.inputs)
        return tuple(interp.lattice.top() for _ in stmt.results)



from bloqade import stim

from .base import codegen


def test_meas():

    @stim.main
    def test_simple_meas():
        stim.MX(p=0.3, targets=(0, 3, 4, 5))

    out = codegen(test_simple_meas)

    assert out.strip() == "MX(0.30000000) 0 3 4 5"



"""Inline QASM dialect.

This dialect allows users to use QASM string as part of a `@qasm2.main` kernel.
"""

import ast
import textwrap
from dataclasses import dataclass

from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.print import Printer
from kirin.exceptions import DialectLoweringError

dialect = ir.Dialect("qasm2.inline")


@dataclass(frozen=True)
class InlineQASMLowering(ir.FromPythonCall):

    def lower(
        self, stmt: type, state: lowering.LoweringState, node: ast.Call
    ) -> lowering.Result:
        from bloqade.qasm2.parse import loads
        from bloqade.qasm2.parse.lowering import LoweringQASM

        if len(node.args) != 1 or node.keywords:
            raise DialectLoweringError("InlineQASM takes 1 positional argument")
        text = node.args[0]
        # 1. string literal
        if isinstance(text, ast.Constant) and isinstance(text.value, str):
            value = text.value
        elif isinstance(text, ast.Name) and isinstance(text.ctx, ast.Load):
            value = state.get_global(text.id).expect(str)
        else:
            raise DialectLoweringError(
                "InlineQASM takes a string literal or global string"
            )

        raw = textwrap.dedent(value)
        qasm_lowering = LoweringQASM(state)
        qasm_lowering.visit(loads(raw))
        return lowering.Result()


# NOTE: this is a dummy statement that won't appear in IR.
# TODO: maybe we should save the string in IR then rewrite?
#       what would be the use case?
@statement(dialect=dialect)
class InlineQASM(ir.Statement):
    name = "text"
    traits = frozenset({InlineQASMLowering()})
    text: str = info.attribute(types.String)

    def __init__(self, text: str) -> None:
        super().__init__(attributes={"text": ir.PyAttr(text)})

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print('"""')
        for line in self.text.splitlines():
            printer.plain_print(line)
            printer.print_newline()
        printer.plain_print('"""')



from kirin import interp

from . import stmts
from .types import RecordResult
from ._dialect import dialect


@dialect.register
class StimAuxMethods(interp.MethodTable):

    @interp.impl(stmts.ConstFloat)
    @interp.impl(stmts.ConstInt)
    @interp.impl(stmts.ConstBool)
    @interp.impl(stmts.ConstStr)
    def const(
        self,
        interpreter: interp.Interpreter,
        frame: interp.Frame,
        stmt: stmts.ConstFloat | stmts.ConstInt | stmts.ConstBool | stmts.ConstStr,
    ):
        return (stmt.value,)

    @interp.impl(stmts.Neg)
    def neg(
        self,
        interpreter: interp.Interpreter,
        frame: interp.Frame,
        stmt: stmts.Neg,
    ):
        return (-frame.get(stmt.operand),)

    @interp.impl(stmts.GetRecord)
    def get_rec(
        self,
        interpreter: interp.Interpreter,
        frame: interp.Frame,
        stmt: stmts.GetRecord,
    ):
        return (RecordResult(value=frame.get(stmt.id)),)



from dataclasses import dataclass

from kirin import ir
from kirin.passes import Pass
from kirin.rewrite import abc, walk, result
from kirin.dialects import py
from bloqade.qasm2.dialects import core


class IndexingDesugarRule(abc.RewriteRule):
    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if isinstance(node, py.indexing.GetItem):
            if node.obj.type.is_subseteq(core.QRegType):
                node.replace_by(core.QRegGet(reg=node.obj, idx=node.index))
                return result.RewriteResult(has_done_something=True)
            elif node.obj.type.is_subseteq(core.CRegType):
                node.replace_by(core.CRegGet(reg=node.obj, idx=node.index))
                return result.RewriteResult(has_done_something=True)

        return result.RewriteResult()


@dataclass
class IndexingDesugarPass(Pass):
    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:

        return walk.Walk(IndexingDesugarRule()).rewrite(mt.code)



from typing import Dict, List, Tuple, Sequence
from dataclasses import field, dataclass

from kirin import ir, types, passes
from bloqade import noise, qasm2
from bloqade.qbraid import schema
from kirin.dialects import func


@ir.dialect_group([func, qasm2.core, qasm2.uop, qasm2.expr, noise.native])
def qbraid_noise(
    self,
):
    fold_pass = passes.Fold(self)
    typeinfer_pass = passes.TypeInfer(self)

    def run_pass(
        method: ir.Method,
        *,
        fold: bool = True,
    ):
        method.verify()

        if fold:
            fold_pass(method)

        typeinfer_pass(method)
        method.code.typecheck()

    return run_pass


@dataclass
class Lowering:
    qubit_list: List[ir.SSAValue] = field(init=False, default_factory=list)
    qubit_id_map: Dict[int, ir.SSAValue] = field(init=False, default_factory=dict)
    bit_id_map: Dict[int, ir.SSAValue] = field(init=False, default_factory=dict)
    block_list: List[ir.Statement] = field(init=False, default_factory=list)

    def lower(
        self,
        sym_name: str,
        noise_model: schema.NoiseModel,
        return_qreg: bool = False,
    ) -> ir.Method:
        """Lower the noise model to a method.

        Args:
            name (str): The name of the method to generate.
            return_qreg (bool): Use the quantum register as the return value.

        Returns:
            Method: The generated kirin method.

        """
        self.process_noise_model(noise_model, return_qreg)
        block = ir.Block(stmts=self.block_list)
        ret_type = qasm2.types.QRegType if return_qreg else qasm2.types.CRegType
        block.args.append_from(types.MethodType[[], ret_type], name=f"{sym_name}_self")
        region = ir.Region(block)
        func_stmt = func.Function(
            sym_name=sym_name,
            signature=func.Signature(inputs=(), output=qasm2.types.QRegType),
            body=region,
        )

        mt = ir.Method(
            mod=None,
            py_func=None,
            sym_name=sym_name,
            dialects=qbraid_noise,
            code=func_stmt,
            arg_names=[],
        )
        qbraid_noise.run_pass(mt)  # type: ignore

        return mt

    def process_noise_model(self, noise_model: schema.NoiseModel, return_qreg: bool):
        num_qubits = self.lower_number(noise_model.num_qubits)

        reg = qasm2.core.QRegNew(num_qubits)
        creg = qasm2.core.CRegNew(num_qubits)
        self.block_list.append(reg)
        self.block_list.append(creg)

        for idx_value, qubit in enumerate(noise_model.all_qubits):
            idx = self.lower_number(idx_value)
            qubit_stmt = qasm2.core.QRegGet(reg.result, idx)
            bit_stmt = qasm2.core.CRegGet(creg.result, idx)

            self.block_list.append(qubit_stmt)
            self.block_list.append(bit_stmt)
            self.qubit_id_map[qubit] = qubit_stmt.result
            self.bit_id_map[qubit] = bit_stmt.result
            self.qubit_list.append(qubit_stmt.result)

        for gate_event in noise_model.gate_events:
            self.process_gate_event(gate_event)

        if return_qreg:
            self.block_list.append(func.Return(reg.result))
        else:
            self.block_list.append(func.Return(creg.result))

    def process_gate_event(self, node: schema.GateEvent):
        self.lower_atom_loss(node.error.survival_prob)

        if isinstance(node.operation, schema.CZ):
            assert isinstance(node.error, schema.CZError), "Only CZError is supported"

            self.process_cz_pauli_error(node.operation.participants, node.error)
            self.lower_cz_gates(node.operation)
        else:

            error = node.error
            assert isinstance(
                error, schema.SingleQubitError
            ), "Only SingleQubitError is supported"
            self.lower_pauli_errors(error.operator_error)

            operation = node.operation
            assert isinstance(
                operation,
                (
                    schema.GlobalW,
                    schema.LocalW,
                    schema.GlobalRz,
                    schema.LocalRz,
                    schema.Measurement,
                ),
            ), f"Only W and Rz gates are supported, found {type(operation)}.__name__"

            if isinstance(operation, schema.GlobalW):
                self.lower_w_gates(
                    tuple(self.qubit_id_map.keys()), operation.theta, operation.phi
                )
            elif isinstance(operation, schema.LocalW):
                self.lower_w_gates(
                    operation.participants, operation.theta, operation.phi
                )
            elif isinstance(operation, schema.GlobalRz):
                self.lower_rz_gates(tuple(self.qubit_id_map.keys()), operation.phi)
            elif isinstance(operation, schema.LocalRz):
                self.lower_rz_gates(operation.participants, operation.phi)
            elif isinstance(operation, schema.Measurement):
                self.lower_measurement(operation)

    def process_cz_pauli_error(
        self,
        participants: Tuple[Tuple[int] | Tuple[int, int], ...],
        node: schema.CZError[schema.PauliErrorModel],
    ):

        storage_error = node.storage_error
        single_error = node.single_error
        entangled_error = node.entangled_error

        assert isinstance(
            storage_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"
        assert isinstance(
            single_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"
        assert isinstance(
            entangled_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"

        self.lower_pauli_errors(storage_error)

        single_error_dict = dict(single_error.errors)
        entangled_error_dict = dict(entangled_error.errors)

        for participant in participants:
            match participant:
                case (qarg_id,):
                    qarg = self.qubit_id_map[qarg_id]
                    px, py, pz = single_error_dict[qarg_id]
                    px_val = self.lower_number(px)
                    py_val = self.lower_number(py)
                    pz_val = self.lower_number(pz)
                    self.block_list.append(
                        noise.native.PauliChannel(
                            px=px_val, py=py_val, pz=pz_val, qarg=qarg
                        )
                    )

                case (qarg1_id, qarg2_id):
                    qarg1 = self.qubit_id_map[qarg1_id]
                    qarg2 = self.qubit_id_map[qarg2_id]
                    # add single qubit errors if
                    # atom is lost during the execution of the CZ gate
                    self.lower_cz_pauli_channel(
                        False,
                        qarg1,
                        qarg2,
                        single_error_dict[qarg1_id],
                        single_error_dict[qarg2_id],
                    )

                    # add entangled errors if
                    # both qubits are active during the execution of the CZ gate
                    self.lower_cz_pauli_channel(
                        True,
                        qarg1,
                        qarg2,
                        entangled_error_dict[qarg1_id],
                        entangled_error_dict[qarg2_id],
                    )

    def lower_cz_gates(self, node: schema.CZ):
        for participant in node.participants:
            if len(participant) != 2:
                continue

            self.block_list.append(
                qasm2.uop.CZ(
                    ctrl=self.qubit_id_map[participant[0]],
                    qarg=self.qubit_id_map[participant[1]],
                )
            )

    def lower_w_gates(self, participants: Sequence[int], theta: float, phi: float):
        for participant in participants:
            self.block_list.append(
                qasm2.uop.UGate(
                    theta=self.lower_full_turns(theta),
                    phi=self.lower_full_turns(phi + 0.5),
                    lam=self.lower_full_turns(-(0.5 + phi)),
                    qarg=self.qubit_id_map[participant],
                )
            )

    def lower_rz_gates(self, participants: Tuple[int, ...], phi: float):
        for participant in participants:
            self.block_list.append(
                qasm2.uop.RZ(
                    theta=self.lower_full_turns(phi),
                    qarg=self.qubit_id_map[participant],
                )
            )

    def lower_pauli_errors(self, operator_error: schema.PauliErrorModel):
        assert isinstance(
            operator_error, schema.PauliErrorModel
        ), "Only PauliErrorModel is supported"

        for qubit_num, (px, py, pz) in operator_error.errors:
            qubit = self.qubit_id_map[qubit_num]
            px_val = self.lower_number(px)
            py_val = self.lower_number(py)
            pz_val = self.lower_number(pz)
            self.block_list.append(
                noise.native.PauliChannel(px=px_val, py=py_val, pz=pz_val, qarg=qubit)
            )

    def lower_measurement(self, operation: schema.Measurement):
        for participant in operation.participants:
            qubit = self.qubit_id_map[participant]
            bit = self.bit_id_map[participant]
            self.block_list.append(qasm2.core.Measure(qarg=qubit, carg=bit))

    def lower_atom_loss(self, survival_probs: Tuple[float, ...]):
        for survival_prob, qubit in zip(survival_probs, self.qubit_list):
            prob = self.lower_number(survival_prob)
            self.block_list.append(noise.native.AtomLossChannel(prob=prob, qarg=qubit))

    def lower_cz_pauli_channel(
        self,
        paired: bool,
        qarg1: ir.SSAValue,
        qarg2: ir.SSAValue,
        p1: Tuple[float, float, float],
        p2: Tuple[float, float, float],
    ):
        px1_val = self.lower_number(p1[0])
        py1_val = self.lower_number(p1[1])
        pz1_val = self.lower_number(p1[2])
        px2_val = self.lower_number(p2[0])
        py2_val = self.lower_number(p2[1])
        pz2_val = self.lower_number(p2[2])

        self.block_list.append(
            noise.native.CZPauliChannel(
                paired=paired,
                px_1=px1_val,
                py_1=py1_val,
                pz_1=pz1_val,
                px_2=px2_val,
                py_2=py2_val,
                pz_2=pz2_val,
                qarg1=qarg1,
                qarg2=qarg2,
            )
        )

    def lower_number(self, value: float | int) -> ir.SSAValue:
        if isinstance(value, int):
            stmt = qasm2.expr.ConstInt(value=value)
        else:
            stmt = qasm2.expr.ConstFloat(value=value)

        self.block_list.append(stmt)
        return stmt.result

    def lower_full_turns(self, value: float) -> ir.SSAValue:
        const_pi = qasm2.expr.ConstPI()
        self.block_list.append(const_pi)
        turns = self.lower_number(2 * value)
        mul = qasm2.expr.Mul(const_pi.result, turns)
        self.block_list.append(mul)
        return mul.result



import bisect
import functools
from typing import Optional

import tqdm
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation

from .base import FieldOfView, GatePainter, quera_color_code
from .runtime.qpustate import QPUStateABC


def animate_qpu_state(
    state: QPUStateABC,
    display_fov: Optional[FieldOfView] = None,
    dilation_rate: float = 0.05,
    fps: int = 30,
    gate_display_dilation: float = 1.0,
    fig_args={},
    save_mpeg: bool = False,
    filename: str = "vqpu_animation",
    start_block: int = 0,
    n_blocks: int | None = None,
):
    """Generate an animation from the QPU state

    Args:
        state (QPUStateABC): The QPU state to animate
        display_fov (Optional[FieldOfView], optional): The field of view to display. Defaults to None. If None, it will use the QPU's field of view.
        dilation_rate (float, optional): The rate at which to dilate the time. Defaults to 0.05.
        fps (int, optional): The frames per second. Defaults to 30.
        gate_display_dilation (float, optional): The rate at which to dilate the gate display. Defaults to 1.0.
        fig_args (dict, optional): The arguments to pass to the matplotlib.pyplot.figure. Defaults to {}.
        save_mpeg (bool, optional): Whether to save the animation as an mpeg. Defaults to False.
        filename (str, optional): The filename to save the mpeg as. Defaults to "vqpu_animation".
        start_block (int, optional): The block to start the animation at. Defaults to 0.
        n_blocks (int | None, optional): The number of blocks to animate. Defaults to None. If None, it will animate all blocks after `start_block`.

    """
    qpu_fov = state.qpu_fov

    if display_fov is None:
        display_fov = qpu_fov

    if start_block >= len(state.block_durations) or start_block < 0:
        raise ValueError("Start block index is out of range")

    if n_blocks is None:
        n_blocks = len(state.block_durations) - start_block

    if n_blocks < 0:
        raise ValueError("Number of block to animate must be non-negative")

    slm_sites = state.get_slm_sites()

    # Scale the figure to different screens and so that the number of SLM sites has the same
    # "area" on screen
    nsites = max([4, len(slm_sites)])
    scale = (
        np.sqrt(44.0 / nsites) * 2.0 * plt.rcParams["figure.dpi"] / 100
    )  # scale the size of the figure

    # figure:
    new_fig_args = {"figsize": (14, 8), **fig_args}
    fig, mpl_axs = plt.subplot_mosaic(
        mosaic=[["Reg", "Info"], ["Reg", "Gate"], ["Reg", "Gate"]],
        gridspec_kw={"width_ratios": [3, 1]},
        **new_fig_args,
    )

    # mpl_axs["Reg"].axis("equal")  # Axis equal must come before axis limits
    mpl_axs["Reg"].set_xlim(left=display_fov.xmin, right=display_fov.xmax)
    mpl_axs["Reg"].set_ylim(bottom=display_fov.ymin, top=display_fov.ymax)
    mpl_axs["Reg"].set(xlabel="x (um)", ylabel="y (um)")
    mpl_axs["Reg"].set_aspect("equal")

    # slm:
    slm_plt_arg = {
        "facecolors": "none",
        "edgecolors": "k",
        "linestyle": "-",
        "s": 80 * scale,
        "alpha": 0.3,
        "linewidth": 0.5 * np.sqrt(scale),
    }
    mpl_axs["Reg"].scatter(
        x=slm_sites[:, 0], y=slm_sites[:, 1], **slm_plt_arg
    )  # this is statically generated, so it will be the background

    # atoms:
    reg_plt_arg = {
        "s": 65 * scale,
        "marker": "o",
        "facecolors": quera_color_code.purple,
        "alpha": 1.0,
    }
    reg_panel = mpl_axs["Reg"]
    reg_scat = reg_panel.scatter([], [], **reg_plt_arg)

    # gates:
    gp = GatePainter(mpl_ax=reg_panel, qpu_fov=qpu_fov, scale=scale)

    # annotate_args = {"fontsize": 8, "ha": "center", "alpha": 0.7, "color": quera_color_code.yellow}
    annotate_args = {
        "fontsize": 6 * np.sqrt(scale),
        "ha": "center",
        "va": "center",
        "alpha": 1.0,
        "color": quera_color_code.yellow,
        "weight": "bold",
    }
    reg_annot_list = [
        reg_panel.annotate(f"{i}", atom_position, **annotate_args)
        for i, atom_position in state.get_atoms_position(time=0.0, include_lost=False)
    ]

    # AODs:
    aod_plot_args = {
        "s": 260 * scale,
        "marker": "+",
        "alpha": 0.7,
        "facecolors": quera_color_code.red,
        "zorder": -100,
        "linewidth": np.sqrt(scale),
    }
    aod_scat = reg_panel.scatter(x=[], y=[], **aod_plot_args)

    aod_h_args = {
        "s": 1e20,
        "marker": "|",
        "alpha": 1.0,
        "color": "#FFE8E9",
        "zorder": -101,
        "linewidth": 0.5 * np.sqrt(scale),
    }
    aod_h_scat = reg_panel.scatter(x=[], y=[], **aod_h_args)
    aod_v_args = {
        "s": 1e20,
        "marker": "_",
        "alpha": 1.0,
        "color": "#FFE8E9",
        "zorder": -101,
        "linewidth": 0.5 * np.sqrt(scale),
    }
    aod_v_scat = reg_panel.scatter(x=[], y=[], **aod_v_args)

    ## Info Panel
    info_text = mpl_axs["Info"].text(x=0.05, y=0.5, s="")
    mpl_axs["Info"].set_xticks([])
    mpl_axs["Info"].set_yticks([])
    mpl_axs["Info"].grid(False)

    ## Event Panel:
    log_text = mpl_axs["Gate"].text(x=0.05, y=0.0, s="", size=6)
    mpl_axs["Gate"].set_xticks([])
    mpl_axs["Gate"].set_yticks([])
    mpl_axs["Gate"].grid(False)

    tstep_mv = 1.0 / (fps * dilation_rate)
    tstep_gate = 1.0 / (fps * dilation_rate * gate_display_dilation)
    blk_t_end = np.cumsum(state.block_durations)

    # determine the dilation part of the timeline, and generate more frame
    chunk_times = []
    curr_t = 0 if start_block == 0 else blk_t_end[start_block - 1]

    for glb_tstart_gate, duration in state.get_gate_events_timing():
        if glb_tstart_gate < curr_t:  # gate start before the current time
            if glb_tstart_gate + duration < curr_t:
                continue
        else:
            dt = glb_tstart_gate - curr_t
            chunk_times.append(np.linspace(curr_t, glb_tstart_gate, int(dt / tstep_mv)))
            curr_t = glb_tstart_gate

        t_gate_end = glb_tstart_gate + duration
        dt = t_gate_end - curr_t
        chunk_times.append(np.linspace(curr_t, t_gate_end, int(dt / tstep_gate)))
        curr_t = t_gate_end

    dt = blk_t_end[-1] - curr_t
    chunk_times.append(np.linspace(curr_t, blk_t_end[-1], int(dt / tstep_mv)))

    times = np.concatenate(chunk_times)

    fig.tight_layout()
    fig.subplots_adjust(wspace=0.1)

    def _update_annotate(loc, idx, annotate_artist):
        new_loc = (loc[0], loc[1] - 0.06)
        annotate_artist.set_position(new_loc)
        txt = f"{idx}"
        annotate_artist.set_text(txt)
        return loc

    def update(frame: int, state: QPUStateABC, times: np.ndarray, blk_t_end: np.array):

        # get positions:

        blk_id = bisect.bisect_left(blk_t_end, times[frame])
        lbl = f"Block: [{blk_id}]\n"
        lbl += f"Block dur: {state.block_durations[blk_id]:.2f} us\n"
        lbl += f"Total elapsed time: {times[frame]:.2f} us"
        info_text.set_text(lbl)

        # update atoms location and annotation
        post = np.array(
            [
                _update_annotate(
                    atom_position,
                    i,
                    reg_annot_list[i],
                )
                for i, atom_position in state.get_atoms_position(
                    times[frame], include_lost=False
                )
            ]
        )
        post = post if post.size > 0 else np.array([(None, None)])
        reg_scat.set_offsets(post)

        # update log event panels
        lost_events = state.get_atoms_lost_info(times[frame])

        # update log gate:
        gate_events = state.get_gate_events(times[frame])
        gate_events_log = [
            f"Gate: {gate.cls_name} @ {t:.6f} (us)\n"
            for t, gate in state.get_gate_events(times[frame])
        ]
        log_text.set_text("".join(lost_events) + "".join(gate_events_log))

        gate_artists = gp.process_gates([gate for _, gate in gate_events])

        # update AODs
        post = state.sample_aod_traps(times[frame]) or [(None, None)]
        aod_scat.set_offsets(post)
        aod_v_scat.set_offsets(post)
        aod_h_scat.set_offsets(post)

        return (
            [reg_scat, info_text, log_text, aod_scat, aod_v_scat, aod_h_scat]
            + reg_annot_list
            + gate_artists
        )

    ani = FuncAnimation(
        fig=fig,
        func=functools.partial(update, state=state, times=times, blk_t_end=blk_t_end),
        frames=len(times),
        interval=tstep_mv,
        blit=True,
        repeat=False,
    )
    if save_mpeg:
        n_frame = len(times)
        pbar = tqdm.tqdm(range(n_frame))

        def p_call_back(i, total_n):
            pbar.update()

        ani.save(
            f"{filename}.mp4", writer="ffmpeg", fps=fps, progress_callback=p_call_back
        )
    else:
        return ani



from kirin.decl import statement

from .base import ControlledTwoQubitGate
from .._dialect import dialect


# Two Qubit Clifford Gates
# ---------------------------------------
@statement(dialect=dialect)
class CX(ControlledTwoQubitGate):
    name = "CX"


@statement(dialect=dialect)
class CY(ControlledTwoQubitGate):
    name = "CY"


@statement(dialect=dialect)
class CZ(ControlledTwoQubitGate):
    name = "CZ"



"""This dialect provides the indexing syntax in Python lowering
for QASM2 dialects. The dialect itself does not contain new statements.

Using this dialect will be conflict with Python semantics provided by
`kirin.dialects.py.binop` and `kirin.dialects.py.indexing` dialects.
"""

import ast

from kirin import ir, types
from kirin.lowering import Result, FromPythonAST, LoweringState
from kirin.exceptions import DialectLoweringError
from bloqade.qasm2.types import BitType, CRegType, QRegType, QubitType
from bloqade.qasm2.dialects import core

dialect = ir.Dialect("qasm2.indexing")


@dialect.register
class QASMCoreLowering(FromPythonAST):
    def lower_Compare(self, state: LoweringState, node: ast.Compare) -> Result:
        lhs = state.visit(node.left).expect_one()
        if len(node.ops) != 1:
            raise DialectLoweringError(
                "only one comparison operator and == is supported for qasm2 lowering"
            )
        rhs = state.visit(node.comparators[0]).expect_one()
        if isinstance(node.ops[0], ast.Eq):
            stmt = core.CRegEq(lhs, rhs)
        else:
            raise DialectLoweringError(
                f"unsupported comparison operator {node.ops[0]} only Eq is supported."
            )

        return Result(state.append_stmt(stmt))

    def lower_Subscript(self, state: LoweringState, node: ast.Subscript) -> Result:
        value = state.visit(node.value).expect_one()
        index = state.visit(node.slice).expect_one()

        if not index.type.is_subseteq(types.Int):
            raise DialectLoweringError(
                f"unsupported subscript index type {index.type},"
                " only integer indices are supported in QASM 2.0"
            )

        if not isinstance(node.ctx, ast.Load):
            raise DialectLoweringError(
                f"unsupported subscript context {node.ctx},"
                " cannot write to subscript in QASM 2.0"
            )

        if value.type.is_subseteq(QRegType):
            stmt = core.QRegGet(reg=value, idx=index)
            stmt.result.type = QubitType
        elif value.type.is_subseteq(CRegType):
            stmt = core.CRegGet(reg=value, idx=index)
            stmt.result.type = BitType
        else:
            raise DialectLoweringError(
                f"unsupported subscript value type {value.type},"
                " only QReg and CReg are supported in QASM 2.0"
            )

        return Result(state.append_stmt(stmt))



from .const import (
    Neg as Neg,
    ConstInt as ConstInt,
    ConstStr as ConstStr,
    ConstBool as ConstBool,
    ConstFloat as ConstFloat,
)
from .annotate import (
    Tick as Tick,
    Detector as Detector,
    GetRecord as GetRecord,
    NewPauliString as NewPauliString,
    ObservableInclude as ObservableInclude,
)



from bloqade import stim
from bloqade.stim.emit import EmitStimMain

interp = EmitStimMain(stim.main)


def test_gates():
    @stim.main
    def test_single_qubit_gates():
        stim.sqrt_z(targets=(0, 1, 2), dagger=False)
        stim.x(targets=(0, 1, 2))
        stim.y(targets=(0, 1), dagger=True)
        stim.z(targets=(1, 2))
        stim.h(targets=(0, 1, 2), dagger=True)
        stim.s(targets=(0, 1, 2), dagger=False)
        stim.s(targets=(0, 1, 2), dagger=True)

    interp.run(test_single_qubit_gates, args=())
    print(interp.get_output())

    @stim.main
    def test_two_qubit_gates():
        stim.swap(targets=(2, 3))

    interp.run(test_two_qubit_gates, args=())
    print(interp.get_output())

    @stim.main
    def test_controlled_two_qubit_gates():
        stim.cx(controls=(0, 1), targets=(2, 3))
        stim.cy(controls=(0, 1), targets=(2, 3), dagger=True)
        stim.cz(controls=(0, 1), targets=(2, 3))

    interp.run(test_controlled_two_qubit_gates, args=())
    print(interp.get_output())

    # @stim.main
    # def test_spp():
    #     pauli_string = stim.PauliString(string=('X', 'Y', 'Z'), targets=(0, 1, 2), flipped=(True, False, True))
    #     stim.spp(targets=(pauli_string, ), dagger=True)

    # interp.run(test_spp, args=())
    # print(interp.get_output())


def test_noise():
    @stim.main
    def test_depolarize():
        stim.depolarize1(p=0.1, targets=(0, 1, 2))
        stim.depolarize2(p=0.1, targets=(0, 1))

    interp.run(test_depolarize, args=())
    print(interp.get_output())

    @stim.main
    def test_pauli_channel():
        stim.pauli_channel1(px=0.01, py=0.01, pz=0.1, targets=(0, 1, 2))
        stim.pauli_channel2(
            pix=0.01,
            piy=0.01,
            piz=0.1,
            pxi=0.01,
            pxx=0.01,
            pxy=0.01,
            pxz=0.1,
            pyi=0.01,
            pyx=0.01,
            pyy=0.01,
            pyz=0.1,
            pzi=0.1,
            pzx=0.1,
            pzy=0.1,
            pzz=0.2,
            targets=(0, 1, 2, 3),
        )

    interp.run(test_pauli_channel, args=())
    print(interp.get_output())

    @stim.main
    def test_pauli_error():
        stim.x_error(p=0.1, targets=(0, 1, 2))
        stim.y_error(p=0.1, targets=(0, 1))
        stim.z_error(p=0.1, targets=(1, 2))

    interp.run(test_pauli_error, args=())
    print(interp.get_output())


def test_collapse():
    @stim.main
    def test_measure():
        stim.mx(p=0.0, targets=(0, 1, 2))
        stim.my(p=0.01, targets=(0, 1))
        stim.mz(p=0.02, targets=(1, 2))
        stim.mzz(p=0.03, targets=(0, 1, 2, 3))
        stim.myy(p=0.04, targets=(0, 1))
        stim.mxx(p=0.05, targets=(1, 2))

    interp.run(test_measure, args=())
    print(interp.get_output())

    @stim.main
    def test_reset():
        stim.rx(targets=(0, 1, 2))
        stim.ry(targets=(0, 1))
        stim.rz(targets=(1, 2))

    interp.run(test_reset, args=())
    print(interp.get_output())


def test_repetition():
    @stim.main
    def test_repetition_memory():
        stim.rz(targets=(0, 1, 2, 3, 4))
        stim.tick()
        stim.depolarize1(p=0.1, targets=(0, 2, 4))
        stim.cx(controls=(0, 2), targets=(1, 3))
        stim.tick()
        stim.cx(controls=(2, 4), targets=(1, 3))
        stim.tick()
        stim.mz(p=0.1, targets=(1, 3))
        stim.detector(coord=(1, 0), targets=(stim.rec(-2),))
        stim.detector(coord=(3, 0), targets=(stim.rec(-1),))
        stim.rz(targets=(1, 3))
        stim.tick()
        stim.depolarize1(p=0.1, targets=(0, 2, 4))
        stim.cx(controls=(0, 2), targets=(1, 3))
        stim.tick()
        stim.cx(controls=(2, 4), targets=(1, 3))
        stim.tick()
        stim.mz(p=0.1, targets=(1, 3))
        stim.detector(coord=(1, 1), targets=(stim.rec(-2), stim.rec(-4)))
        stim.detector(coord=(3, 1), targets=(stim.rec(-1), stim.rec(-3)))
        stim.mz(p=0.1, targets=(0, 2, 4))
        stim.detector(coord=(1, 2), targets=(stim.rec(-2), stim.rec(-3), stim.rec(-5)))
        stim.detector(coord=(3, 2), targets=(stim.rec(-1), stim.rec(-2), stim.rec(-4)))
        stim.observable_include(idx=0, targets=(stim.rec(-1),))

    interp.run(test_repetition_memory, args=())
    print(interp.get_output())



# type: ignore
from lark import Token
from lark.tree import ParseTree

from . import ast


class BuildError(Exception):
    pass


class Build:

    def build(self, tree: ParseTree) -> ast.Node:
        return getattr(self, f"build_{tree.data}", self.build_generic)(tree)

    def build_generic(self, tree: ParseTree) -> ast.Node:
        raise NotImplementedError(f"No build_{tree.data} method")

    def build_mainprogram(self, tree: ParseTree) -> ast.MainProgram:
        header = self.build(tree.children[0])
        return ast.MainProgram(
            header=header,
            statements=[self.build(stmt) for stmt in tree.children[1:]],
        )

    def build_openqasm(self, tree: ParseTree) -> ast.OPENQASM:
        return ast.OPENQASM(version=self.build(tree.children[0]))

    def build_kirin(self, tree: ParseTree) -> ast.Kirin:
        return ast.Kirin(dialects=[self.build_dialect(each) for each in tree.children])

    def build_dialect(self, tree: ParseTree) -> str:
        return ".".join(each.value for each in tree.children)

    def build_version(self, tree: ParseTree) -> ast.Version:
        return ast.Version(
            int(tree.children[0].value),
            int(tree.children[1].value),
        )

    def build_include(self, tree: ParseTree) -> ast.Include:
        return ast.Include(filename=tree.children[0].value[1:-1])

    def build_qreg(self, tree: ParseTree) -> ast.QReg:
        return ast.QReg(name=tree.children[0].value, size=int(tree.children[1].value))

    def build_creg(self, tree: ParseTree) -> ast.CReg:
        return ast.CReg(name=tree.children[0].value, size=int(tree.children[1].value))

    def build_cx_gate(self, tree: ParseTree) -> ast.Gate:
        return ast.CXGate(
            self.build_bit(tree.children[0]), self.build_bit(tree.children[1])
        )

    def build_ugate(self, tree: ParseTree) -> ast.Gate:
        return ast.UGate(
            theta=self.build_expr(tree.children[0]),
            phi=self.build_expr(tree.children[1]),
            lam=self.build_expr(tree.children[2]),
            qarg=self.build_bit(tree.children[3]),
        )

    def build_barrier(self, tree: ParseTree) -> ast.Barrier:
        return ast.Barrier(qargs=self.build(tree.children[0]))

    def build_inst(self, tree: ParseTree) -> ast.Instruction:
        if tree.children[1] is None:
            params = []
        else:
            params = self.build(tree.children[1])
        return ast.Instruction(
            name=ast.Name(tree.children[0].value),
            params=params,
            qargs=self.build(tree.children[2]),
        )

    def build_opaque(self, tree: ParseTree) -> ast.Opaque:
        return ast.Opaque(
            name=tree.children[0].value,
            cparams=self.build(tree.children[1]) if tree.children[1] else [],
            qparams=self.build(tree.children[2]) if tree.children[2] else [],
        )

    def build_gate(self, tree: ParseTree) -> ast.Gate:
        cparams = tree.children[1]
        qparams = tree.children[2]
        return ast.Gate(
            name=tree.children[0].value,
            cparams=self.build(cparams) if cparams else [],
            qparams=self.build(qparams) if qparams else [],
            body=[self.build(stmt) for stmt in tree.children[3:]],
        )

    def build_noise_pauli1(self, tree: ParseTree) -> ast.NoisePAULI1:
        return ast.NoisePAULI1(
            px=self.build_expr(tree.children[0]),
            py=self.build_expr(tree.children[1]),
            pz=self.build_expr(tree.children[2]),
            qarg=self.build_bit(tree.children[3]),
        )

    def build_glob_u_gate(self, tree: ParseTree) -> ast.GlobUGate:
        return ast.GlobUGate(
            [self.build_bit(each) for each in tree.children[3].children],
            self.build_expr(tree.children[0]),
            self.build_expr(tree.children[1]),
            self.build_expr(tree.children[2]),
        )

    def build_para_u_gate(self, tree: ParseTree) -> ast.ParaU3Gate:
        return ast.ParaU3Gate(
            self.build_expr(tree.children[0]),
            self.build_expr(tree.children[1]),
            self.build_expr(tree.children[2]),
            self.build_parallel_body(tree.children[3]),
        )

    def build_para_cz_gate(self, tree: ParseTree) -> ast.ParaCZGate:
        return ast.ParaCZGate(
            self.build_parallel_body(tree.children[0]),
        )

    def build_para_rz_gate(self, tree: ParseTree) -> ast.ParaRZGate:
        return ast.ParaRZGate(
            self.build_expr(tree.children[0]),
            self.build_parallel_body(tree.children[1]),
        )

    def build_parallel_body(self, tree: ParseTree) -> ast.ParallelQArgs:
        return ast.ParallelQArgs([self.build_task_args(each) for each in tree.children])

    def build_task_args(self, tree: ParseTree) -> tuple[ast.Bit | ast.Name, ...]:
        return tuple(self.build(each) for each in tree.children)

    def build_measure(self, tree: ParseTree) -> ast.Measure:
        return ast.Measure(
            qarg=self.build_bit(tree.children[0]),
            carg=self.build_bit(tree.children[1]),
        )

    def build_reset(self, tree: ParseTree) -> ast.Reset:
        return ast.Reset(qarg=self.build_bit(tree.children[0]))

    def build_ifstmt(self, tree: ParseTree) -> ast.IfStmt:
        return ast.IfStmt(
            cond=ast.Cmp(
                lhs=self.build_expr(tree.children[0]),
                rhs=self.build_expr(tree.children[1]),
            ),
            body=self.build(tree.children[2]),
        )

    def build_ifbody(self, tree: ParseTree) -> list[ast.Node]:
        return [self.build(stmt) for stmt in tree.children]

    def build_qparams(self, tree: ParseTree) -> list[ast.Expr]:
        return [param.value for param in tree.children]

    def build_cparams(self, tree: ParseTree) -> list[ast.Expr]:
        return [param.value for param in tree.children]

    def build_qubits(self, tree: ParseTree) -> list[ast.Bit]:
        return [self.build_bit(qubit) for qubit in tree.children]

    def build_bit(self, tree: ParseTree) -> ast.Name | ast.Bit:
        if isinstance(tree, Token) and tree.type == "IDENTIFIER":
            return ast.Name(tree.value)
        else:
            return ast.Bit(
                name=ast.Name(tree.children[0].value), addr=int(tree.children[1].value)
            )

    def build_params(self, tree: ParseTree) -> list[ast.Expr]:
        return [self.build_expr(param) for param in tree.children]

    def build_call(self, tree: ParseTree) -> ast.Call:
        return ast.Call(
            name=tree.children[0].value,
            args=self.build(tree.children[1]),
        )

    def build_arglist(self, tree: ParseTree) -> list[ast.Expr]:
        return [self.build_expr(arg) for arg in tree.children]

    def build_expr(self, tree: ParseTree) -> ast.Expr:
        if isinstance(tree, Token):
            return self._build_token_expr(tree)
        elif tree.data != "expr":
            return self.build(tree)

        if tree.children[1].type == "PLUS":
            lhs = ast.BinOp(
                op="+",
                lhs=self.build_term(tree.children[0]),
                rhs=self.build_term(tree.children[2]),
            )
        elif tree.children[1].type == "MINUS":
            lhs = ast.BinOp(
                op="-",
                lhs=self.build_term(tree.children[0]),
                rhs=self.build_term(tree.children[2]),
            )
        else:
            raise BuildError(f"Unsupported operator {tree.children[1].type}")

        for i in range(3, len(tree.children), 2):
            if tree.children[i].type == "PLUS":
                lhs = ast.BinOp(
                    op="+",
                    lhs=lhs,
                    rhs=self.build_term(tree.children[i + 1]),
                )
            elif tree.children[i].type == "MINUS":
                lhs = ast.BinOp(
                    op="-",
                    lhs=lhs,
                    rhs=self.build_term(tree.children[i + 1]),
                )
            else:
                raise BuildError(f"Unsupported operator {tree.children[i].type}")

        return lhs

    def build_term(self, tree: ParseTree) -> ast.Expr:
        if isinstance(tree, Token):
            return self._build_token_expr(tree)
        elif tree.data != "term":
            return self.build(tree)

        if tree.children[1].type == "TIMES":
            lhs = ast.BinOp(
                op="*",
                lhs=self.build_factor(tree.children[0]),
                rhs=self.build_factor(tree.children[2]),
            )
        elif tree.children[1].type == "DIVIDE":
            lhs = ast.BinOp(
                op="/",
                lhs=self.build_factor(tree.children[0]),
                rhs=self.build_factor(tree.children[2]),
            )
        else:
            raise BuildError(f"Unsupported operator {tree.children[1].type}")

        for i in range(3, len(tree.children), 2):
            if tree.children[i].type == "TIMES":
                lhs = ast.BinOp(
                    op="*",
                    lhs=lhs,
                    rhs=self.build_factor(tree.children[i + 1]),
                )
            elif tree.children[i].type == "DIVIDE":
                lhs = ast.BinOp(
                    op="/",
                    lhs=lhs,
                    rhs=self.build_factor(tree.children[i + 1]),
                )
            else:
                raise BuildError(f"Unsupported operator {tree.children[i].type}")

        return lhs

    def build_factor(self, tree: ParseTree) -> ast.Expr:
        if isinstance(tree, Token):
            return self._build_token_expr(tree)
        elif tree.data != "factor":
            return self.build(tree)

        if tree.children[0].type == "PLUS":
            return ast.UnaryOp(op="+", operand=self.build_expr(tree.children[1]))
        elif tree.children[0].type == "MINUS":
            return ast.UnaryOp(op="-", operand=self.build_expr(tree.children[1]))
        else:
            raise BuildError(f"Unsupported operator {tree.children[0].type}")

    def _build_token_expr(self, token: Token) -> ast.Node:
        if token.type == "NUMBER":
            return ast.Number(float(token.value))
        elif token.type == "IDENTIFIER":
            return ast.Name(token.value)
        elif token.type == "PI":
            return ast.Pi()
        elif token.type == "INT":
            return ast.Number(int(token.value))
        elif token.type == "FLOAT":
            return ast.Number(float(token.value))
        else:
            raise BuildError(f"Unsupported token type {token.type}")



import json

from bloqade.qbraid import schema


def get_noise_model():
    single_qubit_error = schema.SingleQubitError(
        survival_prob=(survival_prob := tuple(1 for _ in range(10))),
        operator_error=(pauli_error_model := schema.PauliErrorModel()),
    )
    cz_error = schema.CZError(
        survival_prob=survival_prob,
        entangled_error=pauli_error_model,
        single_error=pauli_error_model,
        storage_error=pauli_error_model,
    )
    return schema.NoiseModel(
        all_qubits=(all_qubits := tuple(range(10))),
        gate_events=[
            schema.GateEvent(
                operation=schema.LocalW(
                    participants=(0,),
                    theta=0.25,
                    phi=0,
                ),
                error=single_qubit_error,
            ),
            schema.GateEvent(
                operation=schema.CZ(participants=((0, 1),)),
                error=cz_error,
            ),
            schema.GateEvent(
                operation=schema.CZ(participants=((0, 2),)),
                error=cz_error,
            ),
            schema.GateEvent(
                operation=schema.Measurement(
                    participants=all_qubits,
                ),
                error=single_qubit_error,
            ),
        ],
    )


def test_serialization():
    circuit = get_noise_model()

    circuit_json = circuit.model_dump_json()
    deserialized_circuit = schema.NoiseModel(**json.loads(circuit_json))

    assert circuit == deserialized_circuit



from kirin import ir, types
from kirin.decl import info, statement
from bloqade.qasm2.types import QubitType

from ._dialect import dialect


@statement(dialect=dialect)
class PauliChannel(ir.Statement):
    name = "native.pauli_channel"

    traits = frozenset({ir.FromPythonCall()})

    px: ir.SSAValue = info.argument(type=types.Float)
    py: ir.SSAValue = info.argument(type=types.Float)
    pz: ir.SSAValue = info.argument(type=types.Float)
    qarg: ir.SSAValue = info.argument(type=QubitType)


@statement(dialect=dialect)
class CZPauliChannel(ir.Statement):
    name = "native.pauli_channel.cz_pauli_channel"

    traits = frozenset({ir.FromPythonCall()})

    paired: bool = info.attribute()
    px_1: ir.SSAValue = info.argument(type=types.Float)
    py_1: ir.SSAValue = info.argument(type=types.Float)
    pz_1: ir.SSAValue = info.argument(type=types.Float)
    px_2: ir.SSAValue = info.argument(type=types.Float)
    py_2: ir.SSAValue = info.argument(type=types.Float)
    pz_2: ir.SSAValue = info.argument(type=types.Float)
    qarg1: ir.SSAValue = info.argument(type=QubitType)
    qarg2: ir.SSAValue = info.argument(type=QubitType)


@statement(dialect=dialect)
class AtomLossChannel(ir.Statement):
    name = "native.atom_loss_channel"

    traits = frozenset({ir.FromPythonCall()})

    prob: ir.SSAValue = info.argument(type=types.Float)
    qarg: ir.SSAValue = info.argument(type=QubitType)



from bloqade.noise import native
from kirin.lowering import wraps
from bloqade.qasm2.types import Qubit


@wraps(native.AtomLossChannel)
def atom_loss_channel(prob: float, qarg: Qubit) -> None: ...


@wraps(native.PauliChannel)
def pauli_channel(px: float, py: float, pz: float, qarg: Qubit) -> None: ...


@wraps(native.CZPauliChannel)
def cz_pauli_channel(
    px_1: float,
    py_1: float,
    pz_1: float,
    px_2: float,
    py_2: float,
    pz_2: float,
    qarg1: Qubit,
    qarg2: Qubit,
    *,
    paired: bool,
) -> None: ...



from kirin.emit import EmitStrFrame
from kirin.interp import MethodTable, impl
from bloqade.stim.emit.stim import EmitStimMain

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.stim")
class EmitStimAuxMethods(MethodTable):

    @impl(stmts.ConstInt)
    def const_int(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.ConstInt):

        out: str = f"{stmt.value}"

        return (out,)

    @impl(stmts.ConstFloat)
    def const_float(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.ConstFloat
    ):

        out: str = f"{stmt.value:.8f}"

        return (out,)

    @impl(stmts.ConstBool)
    def const_bool(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.ConstBool
    ):
        out: str = "!" if stmt.value else ""

        return (out,)

    @impl(stmts.ConstStr)
    def const_str(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.ConstBool):

        return (stmt.value,)

    @impl(stmts.Neg)
    def neg(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.Neg):

        operand: str = frame.get(stmt.operand)

        return ("-" + operand,)

    @impl(stmts.GetRecord)
    def get_rec(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.GetRecord):

        id: str = frame.get(stmt.id)
        out: str = f"rec[{id}]"

        return (out,)

    @impl(stmts.Tick)
    def tick(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.Tick):

        emit.writeln(frame, "TICK")

        return ()

    @impl(stmts.Detector)
    def detector(self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.Detector):

        coords: tuple[str, ...] = frame.get_values(stmt.coord)
        targets: tuple[str, ...] = frame.get_values(stmt.targets)

        coord_str: str = ", ".join(coords)
        target_str: str = " ".join(targets)
        emit.writeln(frame, f"DETECTOR({coord_str}) {target_str}")

        return ()

    @impl(stmts.ObservableInclude)
    def obs_include(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.ObservableInclude
    ):

        idx: str = frame.get(stmt.idx)
        targets: tuple[str, ...] = frame.get_values(stmt.targets)

        target_str: str = " ".join(targets)
        emit.writeln(frame, f"OBSERVABLE_INCLUDE({idx}) {target_str}")

        return ()

    @impl(stmts.NewPauliString)
    def new_paulistr(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.NewPauliString
    ):

        string: tuple[str, ...] = frame.get_values(stmt.string)
        flipped: tuple[str, ...] = frame.get_values(stmt.flipped)
        targets: tuple[str, ...] = frame.get_values(stmt.targets)

        out = "*".join(
            f"{flp}{base}{tgt}" for flp, base, tgt in zip(flipped, string, targets)
        )

        return (out,)



from kirin import ir
from kirin.passes import Fold, TypeInfer
from kirin.dialects import func, lowering

from .dialects import aux, gate, noise, collapse


@ir.dialect_group([noise, gate, aux, collapse, func, lowering.func, lowering.call])
def main(self):
    typeinfer_pass = TypeInfer(self)
    fold_pass = Fold(self)

    def run_pass(
        mt: ir.Method,
        *,
        typeinfer: bool = False,
        fold: bool = True,
    ) -> None:

        if typeinfer:
            typeinfer_pass(mt)

        if fold:
            fold_pass(mt)

    return run_pass



from kirin import ir, types
from kirin.decl import info, statement

from .._dialect import dialect


@statement
class Reset(ir.Statement):
    name = "reset"
    traits = frozenset({ir.FromPythonCall()})
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class RZ(Reset):
    name = "RZ"


@statement(dialect=dialect)
class RY(Reset):
    name = "RY"


@statement(dialect=dialect)
class RX(Reset):
    name = "RX"



"""QASM2 extension for parallel execution of gates."""

from typing import Any

from kirin.dialects import ilist
from kirin.lowering import wraps

from .types import Qubit
from .dialects import parallel


@wraps(parallel.CZ)
def cz(
    ctrls: ilist.IList[Qubit, Any] | list, qargs: ilist.IList[Qubit, Any] | list
) -> None:
    """Apply a controlled-Z gate to input qubits in parallel.

    Args:
        ctrls (IList[Qubit] | list[Qubit]): The control qubits.
        qargs (IList[Qubit] | list[Qubit]): The target qubits.

    """


@wraps(parallel.UGate)
def u(
    qargs: ilist.IList[Qubit, Any] | list, theta: float, phi: float, lam: float
) -> None:
    """Apply a U gate to input qubits in parallel.

    Args:
        qargs (IList[Qubit] | list[Qubit]): The target qubits.
        theta (float): The angle theta.
        phi (float): The angle phi.
        lam (float): The angle lam.

    """


@wraps(parallel.RZ)
def rz(qargs: ilist.IList[Qubit, Any] | list, theta: float) -> None:
    """Apply a RZ gate to input qubits in parallel.

    Args:
        qargs (IList[Qubit] | list[Qubit]): The target qubits.
        theta (float): The angle theta.

    """



"""
Passes for converting parallel gates into multiple single gates as well as
converting multiple single gates to parallel gates.
"""

from typing import Type
from dataclasses import dataclass

from kirin import ir
from kirin.rewrite import dce, walk, chain, result, fixpoint
from bloqade.analysis import address, schedule
from kirin.passes.abc import Pass
from bloqade.qasm2.rewrite import (
    MergePolicyABC,
    ParallelToUOpRule,
    UOpToParallelRule,
    SimpleGreedyMergePolicy,
)


@dataclass
class ParallelToUOp(Pass):
    """Pass to convert parallel gates into single gates.

    This pass rewrites any parallel gates from the `qasm2.parallel` dialect into multiple
    single gates in the `qasm2.uop` dialect, bringing the program closer to
    conforming to standard QASM2 syntax.

    ## Usage Examples
    ```
    # Define kernel
    @qasm2.extended
    def main():
        q = qasm2.qreg(4)

        qasm2.parallel.cz(ctrls=[q[0], q[2]], qargs=[q[1], q[3]])

    # Run rewrite
    ParallelToUOp(main.dialects)(main)
    ```

    The `qasm2.parallel.cz` statement has been rewritten to individual gates:

    ```
    qasm2.uop.cz(ctrl=q[0], qarg=q[1])
    qasm2.uop.cz(ctrl=q[2], qarg=q[3])
    ```

    """

    def generate_rule(self, mt: ir.Method) -> ParallelToUOpRule:
        frame, _ = address.AddressAnalysis(mt.dialects).run_analysis(mt)

        id_map = {}

        # GOAL: Get the ssa value for the first reference of each qubit.
        for ssa, addr in frame.entries.items():
            if not isinstance(addr, address.AddressQubit):
                # skip any stmts that are not qubits
                continue

            # get qubit id from analysis result
            qubit_id = addr.data

            # check if id has already been found
            # if so, skip this ssa value
            if qubit_id in id_map:
                continue

            id_map[qubit_id] = ssa

        return ParallelToUOpRule(id_map=id_map, address_analysis=frame.entries)

    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:
        return fixpoint.Fixpoint(
            chain.Chain(
                walk.Walk(self.generate_rule(mt)), walk.Walk(dce.DeadCodeElimination())
            )
        ).rewrite(mt.code)


@dataclass
class UOpToParallel(Pass):
    """Pass to convert single gates into parallel gates.

    This pass looks for single gates from the `qasm2.uop` dialect that can be combined
    into parallel gates from the `qasm2.parallel` dialect and performs a rewrite to do so.

    ## Usage Examples
    ```
    # Define kernel
    @qasm2.main
    def test():
        q = qasm2.qreg(4)

        theta = 0.1
        phi = 0.2
        lam = 0.3

        qasm2.u(q[1], theta, phi, lam)
        qasm2.u(q[3], theta, phi, lam)
        qasm2.cx(q[1], q[3])
        qasm2.u(q[2], theta, phi, lam)
        qasm2.u(q[0], theta, phi, lam)
        qasm2.cx(q[0], q[2])

    # Run rewrite
    UOpToParallel(main.dialects)(main)
    ```

    The individual `qasm2.u` statements have now been combined
    into a single `qasm2.parallel.u` statement.

    ```
    qasm2.parallel.u(qargs = [q[0], q[1], q[2], q[3]], theta, phi, lam)
    qasm2.uop.CX(q[1], q[3])
    qasm2.uop.CX(q[0], q[2])
    ```

    """

    merge_policy_type: Type[MergePolicyABC] = SimpleGreedyMergePolicy

    def generate_rule(self, mt: ir.Method):
        frame, _ = address.AddressAnalysis(mt.dialects).run_analysis(mt)
        dags = schedule.DagScheduleAnalysis(
            mt.dialects, address_analysis=frame.entries
        ).get_dags(mt)

        return UOpToParallelRule(
            {
                block: self.merge_policy_type.from_analysis(dag, frame.entries)
                for block, dag in dags.items()
            }
        )

    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:
        return fixpoint.Fixpoint(
            chain.Chain(
                walk.Walk(self.generate_rule(mt)), walk.Walk(dce.DeadCodeElimination())
            )
        ).rewrite(mt.code)



import dataclasses
from typing import Any, Dict, Tuple

from .ppoly import PPoly


@dataclasses.dataclass(frozen=True)
class AODMoveEvent:
    time: float
    duration: float
    x: PPoly
    y: PPoly

    def __post_init__(self):
        assert isinstance(self.x, PPoly), "x must be a PPoly instance"
        assert isinstance(self.y, PPoly), "y must be a PPoly instance"

    def sample(self, t: float) -> Tuple[float, float]:
        return (self.x(t - self.time), self.y(t - self.time))

    def to_json(self) -> Dict[str, Any]:
        return {
            "time": self.time,
            "duration": self.duration,
            "x": self.x.to_json(),
            "y": self.y.to_json(),
        }

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "AODMoveEvent":
        return cls(
            json_dict["time"],
            json_dict["duration"],
            PPoly.from_json(json_dict["x"]),
            PPoly.from_json(json_dict["y"]),
        )



from kirin import ir, types
from kirin.decl import info, statement

from .._dialect import dialect
from ...aux.types import PauliStringType


@statement(dialect=dialect)
class PPMeasurement(ir.Statement):
    name = "MPP"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    """probability of noise introduced by measurement. For example 0.01 means 1% the measurement will be flipped"""
    targets: tuple[ir.SSAValue, ...] = info.argument(PauliStringType)



from bloqade import stim

from .base import codegen


def test_spp():

    @stim.main
    def test_spp_main():
        stim.SPP(
            targets=(
                stim.NewPauliString(
                    string=("X", "X", "Z"),
                    flipped=(True, False, False),
                    targets=(0, 1, 2),
                ),
                stim.NewPauliString(
                    string=("Y", "X", "Y"),
                    flipped=(False, False, True),
                    targets=(3, 4, 5),
                ),
            ),
            dagger=False,
        )

    test_spp_main.print()
    out = codegen(test_spp_main)
    assert out.strip() == "SPP !X0*X1*Z2 Y3*X4*!Y5"


test_spp()



from typing import Any

from kirin import ir, types, interp
from kirin.decl import info, statement
from kirin.analysis import ForwardFrame
from kirin.dialects import ilist
from bloqade.qasm2.parse import ast
from bloqade.qasm2.types import QubitType
from bloqade.qasm2.emit.gate import EmitQASM2Gate, EmitQASM2Frame
from bloqade.analysis.schedule import DagScheduleAnalysis

dialect = ir.Dialect("qasm2.parallel")

N = types.TypeVar("N")


@statement(dialect=dialect)
class CZ(ir.Statement):
    name = "cz"
    traits = frozenset({ir.FromPythonCall()})
    ctrls: ir.SSAValue = info.argument(ilist.IListType[QubitType, N])
    qargs: ir.SSAValue = info.argument(ilist.IListType[QubitType, N])


@statement(dialect=dialect)
class UGate(ir.Statement):
    name = "u"
    traits = frozenset({ir.FromPythonCall()})
    qargs: ir.SSAValue = info.argument(ilist.IListType[QubitType])
    theta: ir.SSAValue = info.argument(types.Float)
    phi: ir.SSAValue = info.argument(types.Float)
    lam: ir.SSAValue = info.argument(types.Float)


@statement(dialect=dialect)
class RZ(ir.Statement):
    name = "rz"
    traits = frozenset({ir.FromPythonCall()})
    qargs: ir.SSAValue = info.argument(ilist.IListType[QubitType])
    theta: ir.SSAValue = info.argument(types.Float)


@dialect.register(key="emit.qasm2.gate")
class Parallel(interp.MethodTable):

    def _emit_parallel_qargs(
        self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, qargs: ir.SSAValue
    ):
        qargs_: ilist.IList[ast.Node, Any] = frame.get(qargs)  # type: ignore
        return [(emit.assert_node((ast.Name, ast.Bit), qarg),) for qarg in qargs_]

    @interp.impl(UGate)
    def ugate(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: UGate):
        qargs = self._emit_parallel_qargs(emit, frame, stmt.qargs)
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        phi = emit.assert_node(ast.Expr, frame.get(stmt.phi))
        lam = emit.assert_node(ast.Expr, frame.get(stmt.lam))
        frame.body.append(
            ast.ParaU3Gate(
                theta=theta, phi=phi, lam=lam, qargs=ast.ParallelQArgs(qargs=qargs)
            )
        )
        return ()

    @interp.impl(RZ)
    def rz(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: RZ):
        qargs = self._emit_parallel_qargs(emit, frame, stmt.qargs)
        theta = emit.assert_node(ast.Expr, frame.get(stmt.theta))
        frame.body.append(
            ast.ParaRZGate(theta=theta, qargs=ast.ParallelQArgs(qargs=qargs))
        )
        return ()

    @interp.impl(CZ)
    def cz(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: CZ):
        ctrls = self._emit_parallel_qargs(emit, frame, stmt.ctrls)
        qargs = self._emit_parallel_qargs(emit, frame, stmt.qargs)
        frame.body.append(
            ast.ParaCZGate(
                qargs=ast.ParallelQArgs(
                    qargs=[ctrl + qarg for ctrl, qarg in zip(ctrls, qargs)]
                )
            )
        )
        return ()


@dialect.register(key="qasm2.dag")
class ParallelDag(interp.MethodTable):

    @interp.impl(CZ)
    def parallel_cz(self, interp: DagScheduleAnalysis, frame: ForwardFrame, stmt: CZ):
        interp.update_dag(stmt, [stmt.qargs, stmt.ctrls])
        return ()

    @interp.impl(UGate)
    def parallel_ugate(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: UGate,
    ):
        interp.update_dag(stmt, [stmt.qargs])
        return ()

    @interp.impl(RZ)
    def parallel_rz(self, interp: DagScheduleAnalysis, frame: ForwardFrame, stmt: RZ):
        interp.update_dag(stmt, [stmt.qargs])
        return ()



import dataclasses
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

import numpy as np

from .aod import AODMoveEvent
from ..base import FieldOfView
from .atoms import AtomTrajectory
from ..gate_event import GateEvent


@dataclasses.dataclass
class QPUStateABC(ABC):

    block_durations: List[float] = dataclasses.field(default_factory=list)
    gate_events: List[Tuple[float, GateEvent]] = dataclasses.field(default_factory=list)
    qpu_fov: FieldOfView = dataclasses.field(default_factory=FieldOfView)

    def __post_init__(self):
        assert all(
            isinstance(gate, GateEvent) for _, gate in self.gate_events
        ), "All gate events must be GateEvent instances"

    @abstractmethod
    def get_slm_sites(self) -> np.ndarray: ...

    @abstractmethod
    def sample_aod_traps(self, time: float) -> List[Tuple[float, float]]: ...

    @abstractmethod
    def get_atoms_lost_info(self, time: float) -> List[str]: ...

    @abstractmethod
    def get_atoms_position(
        self, time: float, include_lost: bool = False
    ) -> List[Tuple[int, Tuple[float, float]]]: ...

    def get_gate_events_timing(self) -> List[Tuple[float, float]]:
        # return [t_start, duration]
        return [(t, gate.duration) for t, gate in self.gate_events]

    def get_gate_events(self, time: float) -> List[Tuple[float, GateEvent]]:
        out = []
        if self.gate_events:
            for t, gate in self.gate_events:
                if t > time:
                    break
                if t <= time < t + gate.duration:
                    out.append((t, gate))
        return out


@dataclasses.dataclass
class AnimateQPUState(QPUStateABC):
    atoms: List[AtomTrajectory] = dataclasses.field(default_factory=list)
    slm_zone: List[Tuple[float, float]] = dataclasses.field(default_factory=list)
    aod_moves: List[AODMoveEvent] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        assert all(
            isinstance(atom, AtomTrajectory) for atom in self.atoms
        ), "All atoms must be AtomTrajectory instances"
        assert all(
            isinstance(aod, AODMoveEvent) for aod in self.aod_moves
        ), "All AOD moves must be AODMoveEvent instances"
        assert all(
            isinstance(site, tuple) and len(site) == 2 for site in self.slm_zone
        ), "All SLM sites must be tuples of length 2"

    def get_slm_sites(self):
        return np.array(self.slm_zone)

    def sample_aod_traps(self, time: float) -> List[Tuple[float, float]]:
        out = []
        for aod_trap in self.aod_moves:
            if aod_trap.time <= time < aod_trap.time + aod_trap.duration:
                out.append(aod_trap.sample(time))

        return out

    def get_atoms_lost_info(self, time: float) -> List[str]:
        return [
            f"Lost: {atom.id} @{atom.events[-1][0]:.3f} (us)" + "\n"
            for atom in self.atoms
            if atom.is_lost(time)
        ]

    def get_atoms_position(
        self, time: float, include_lost: bool = False
    ) -> List[Tuple[int, Tuple[float, float]]]:
        return [
            (atom.id, atom.position(time))
            for atom in self.atoms
            if not atom.is_lost(time) or include_lost
        ]

    def to_json(self) -> Dict[str, Any]:
        return {
            "block_durations": self.block_durations,
            "gate_events": [(t, gate.to_json()) for t, gate in self.gate_events],
            "qpu_fov": self.qpu_fov.to_json(),
            "atoms": [atom.to_json() for atom in self.atoms],
            "slm_zone": self.slm_zone,
            "aod_moves": [aod.to_json() for aod in self.aod_moves],
        }

    @classmethod
    def from_json(cls, json_dict: Dict[str, Any]) -> "AnimateQPUState":
        return cls(
            block_durations=json_dict["block_durations"],
            gate_events=[
                (t, GateEvent.from_json(gate)) for t, gate in json_dict["gate_events"]
            ],
            qpu_fov=FieldOfView.from_json(json_dict["qpu_fov"]),
            atoms=list(map(AtomTrajectory.from_json, json_dict["atoms"])),
            slm_zone=list(map(tuple, json_dict["slm_zone"])),
            aod_moves=list(map(AODMoveEvent.from_json, json_dict["aod_moves"])),
        )



from kirin import interp
from bloqade.qasm2.parse import ast
from bloqade.qasm2.emit.main import EmitQASM2Main, EmitQASM2Frame

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.qasm2.main")
class Core(interp.MethodTable):

    @interp.impl(stmts.CRegNew)
    def emit_creg_new(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: stmts.CRegNew
    ):
        n_bits = emit.assert_node(ast.Number, frame.get(stmt.n_bits))
        # check if its int first, because Int.is_integer() is added for >=3.12
        assert isinstance(n_bits.value, int), "expected integer"
        name = emit.ssa_id[stmt.result]
        frame.body.append(ast.CReg(name=name, size=int(n_bits.value)))
        return (ast.Name(name),)

    @interp.impl(stmts.QRegNew)
    def emit_qreg_new(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: stmts.QRegNew
    ):
        n_bits = emit.assert_node(ast.Number, frame.get(stmt.n_qubits))
        assert isinstance(n_bits.value, int), "expected integer"
        name = emit.ssa_id[stmt.result]
        frame.body.append(ast.QReg(name=name, size=int(n_bits.value)))
        return (ast.Name(name),)

    @interp.impl(stmts.Reset)
    def emit_reset(self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: stmts.Reset):
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        frame.body.append(ast.Reset(qarg=qarg))
        return ()

    @interp.impl(stmts.Measure)
    def emit_measure(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: stmts.Measure
    ):
        qarg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.qarg))
        carg = emit.assert_node((ast.Bit, ast.Name), frame.get(stmt.carg))
        frame.body.append(ast.Measure(qarg=qarg, carg=carg))
        return ()

    @interp.impl(stmts.CRegEq)
    def emit_creg_eq(
        self, emit: EmitQASM2Main, frame: EmitQASM2Frame, stmt: stmts.CRegEq
    ):
        lhs = emit.assert_node(ast.Expr, frame.get(stmt.lhs))
        rhs = emit.assert_node(ast.Expr, frame.get(stmt.rhs))
        return (ast.Cmp(lhs=lhs, rhs=rhs),)

    @interp.impl(stmts.CRegGet)
    @interp.impl(stmts.QRegGet)
    def emit_qreg_get(
        self,
        emit: EmitQASM2Main,
        frame: EmitQASM2Frame,
        stmt: stmts.QRegGet | stmts.CRegGet,
    ):
        reg = emit.assert_node(ast.Name, frame.get(stmt.reg))
        idx = emit.assert_node(ast.Number, frame.get(stmt.idx))
        assert isinstance(idx.value, int)
        return (ast.Bit(reg, int(idx.value)),)



import ast

from kirin import ir, types
from kirin.lowering import Result, FromPythonAST, LoweringState
from kirin.exceptions import DialectLoweringError

from . import stmts
from ._dialect import dialect


@dialect.register
class QASMUopLowering(FromPythonAST):

    def lower_Name(self, state: LoweringState, node: ast.Name) -> Result:
        name = node.id
        if isinstance(node.ctx, ast.Load):
            value = state.current_frame.get(name)
            if value is None:
                raise DialectLoweringError(f"{name} is not defined")
            return Result(value)
        elif isinstance(node.ctx, ast.Store):
            raise DialectLoweringError("unhandled store operation")
        else:  # Del
            raise DialectLoweringError("unhandled del operation")

    def lower_Assign(self, state: LoweringState, node: ast.Assign) -> Result:
        # NOTE: QASM only expects one value on right hand side
        rhs = state.visit(node.value).expect_one()
        current_frame = state.current_frame
        match node:
            case ast.Assign(targets=[ast.Name(lhs_name, ast.Store())], value=_):
                rhs.name = lhs_name
                current_frame.defs[lhs_name] = rhs
            case _:
                target_syntax = ", ".join(
                    ast.unparse(target) for target in node.targets
                )
                raise DialectLoweringError(f"unsupported target syntax {target_syntax}")
        return Result()  # python assign does not have value

    def lower_Expr(self, state: LoweringState, node: ast.Expr) -> Result:
        return state.visit(node.value)

    def lower_Constant(self, state: LoweringState, node: ast.Constant) -> Result:
        if isinstance(node.value, int):
            stmt = stmts.ConstInt(value=node.value)
            return Result(state.append_stmt(stmt))
        elif isinstance(node.value, float):
            stmt = stmts.ConstFloat(value=node.value)
            return Result(state.append_stmt(stmt))
        else:
            raise DialectLoweringError(
                f"unsupported QASM 2.0 constant type {type(node.value)}"
            )

    def lower_BinOp(self, state: LoweringState, node: ast.BinOp) -> Result:
        lhs = state.visit(node.left).expect_one()
        rhs = state.visit(node.right).expect_one()
        if isinstance(node.op, ast.Add):
            stmt = stmts.Add(lhs, rhs)
        elif isinstance(node.op, ast.Sub):
            stmt = stmts.Sub(lhs, rhs)
        elif isinstance(node.op, ast.Mult):
            stmt = stmts.Mul(lhs, rhs)
        elif isinstance(node.op, ast.Div):
            stmt = stmts.Div(lhs, rhs)
        elif isinstance(node.op, ast.Pow):
            stmt = stmts.Pow(lhs, rhs)
        else:
            raise DialectLoweringError(f"unsupported QASM 2.0 binop {node.op}")
        stmt.result.type = self.__promote_binop_type(lhs, rhs)
        return Result(state.append_stmt(stmt))

    def __promote_binop_type(
        self, lhs: ir.SSAValue, rhs: ir.SSAValue
    ) -> types.TypeAttribute:
        if lhs.type.is_subseteq(types.Float) or rhs.type.is_subseteq(types.Float):
            return types.Float
        return types.Int

    def lower_UnaryOp(self, state: LoweringState, node: ast.UnaryOp) -> Result:
        if isinstance(node.op, ast.USub):
            value = state.visit(node.operand).expect_one()
            stmt = stmts.Neg(value)
            return Result(state.append_stmt(stmt))
        elif isinstance(node.op, ast.UAdd):
            return state.visit(node.operand)
        else:
            raise DialectLoweringError(f"unsupported QASM 2.0 unaryop {node.op}")



import os
import pathlib

from bloqade.qasm2.parse import loads, spprint, loadfile
from bloqade.qasm2.parse.parser import qasm2_parser as lark_parser


def roundtrip(file):
    ast1 = loadfile(os.path.join(os.path.dirname(__file__), "programs", file))
    ast2 = loads(spprint(ast1))
    return ast1 == ast2


def test_roundtrip():
    path = pathlib.Path(__file__).parent / "programs"
    for file in path.glob("*.qasm"):
        assert roundtrip(file.name), f"Failed roundtrip for {file}"


if __name__ == "__main__":
    filepath = os.path.join(os.path.dirname(__file__), "programs", "global.qasm")
    print(filepath)
    with open(filepath) as f:
        qasm_str = f.read()

    parse_tree = lark_parser.parse(qasm_str)  # raw parsing seems to be fine
    print(parse_tree.pretty())

    ast = loads(qasm_str)
    print(spprint(ast))



from bloqade import stim

from .base import codegen


def test_detector():

    @stim.main
    def test_simple_cx():
        stim.Detector(coord=(1, 2, 3), targets=(stim.GetRecord(-3), stim.GetRecord(-1)))

    out = codegen(test_simple_cx)

    assert out.strip() == "DETECTOR(1, 2, 3) rec[-3] rec[-1]"


test_detector()



from __future__ import annotations

from typing import Union, Literal
from dataclasses import dataclass


@dataclass
class Node:
    pass


@dataclass
class Header(Node):
    pass


@dataclass
class Kirin(Header):
    dialects: list[str]


@dataclass
class OPENQASM(Header):
    version: Version


@dataclass
class MainProgram(Node):
    header: Kirin | OPENQASM
    statements: list[Statement]


@dataclass
class Version(Node):
    major: int
    minor: int


@dataclass
class Statement(Node):
    pass


@dataclass
class Comment(Statement):
    text: str


@dataclass
class QReg(Statement):
    name: str
    size: int


@dataclass
class CReg(Statement):
    name: str
    size: int


@dataclass
class Gate(Statement):
    name: str
    cparams: list[str]
    qparams: list[str]
    body: list[UOp | Barrier]


@dataclass
class Opaque(Statement):
    name: str
    cparams: list[Expr]
    qparams: list[Bit | Name]


@dataclass
class QOp(Statement):
    pass


@dataclass
class IfStmt(Statement):
    cond: Cmp
    body: list[QOp]


@dataclass
class Barrier(Statement):
    qargs: list[Bit | Name]


@dataclass
class Include(Statement):
    filename: str


@dataclass
class Measure(QOp):
    qarg: Bit | Name
    carg: Bit | Name


@dataclass
class Reset(QOp):
    qarg: Bit | Name


@dataclass
class UOp(QOp):
    pass


@dataclass
class Instruction(UOp):
    name: Name
    params: list[Expr]
    qargs: list[Bit | Name]


@dataclass
class UGate(UOp):
    theta: Expr
    phi: Expr
    lam: Expr
    qarg: Bit | Name


@dataclass
class CXGate(UOp):
    ctrl: Bit | Name
    qarg: Bit | Name


@dataclass
class Extension(UOp):
    pass


@dataclass
class ParallelQArgs(Extension):
    qargs: list[tuple[Bit | Name, ...]]

    def __iter__(self):
        return iter(self.qargs)


@dataclass
class GlobUGate(Extension):
    registers: list[Name]
    theta: Expr
    phi: Expr
    lam: Expr


@dataclass
class NoisePAULI1(Extension):
    px: Expr
    py: Expr
    pz: Expr
    qarg: Bit | Name


@dataclass
class ParaU3Gate(Extension):
    theta: Expr
    phi: Expr
    lam: Expr
    qargs: ParallelQArgs


@dataclass
class ParaRZGate(Extension):
    theta: Expr
    qargs: ParallelQArgs


@dataclass
class ParaCZGate(Extension):
    qargs: ParallelQArgs


@dataclass
class Expr(Node):
    pass


@dataclass
class BinOp(Expr):
    op: Literal["+", "-", "*", "/", "^"]
    lhs: Expr
    rhs: Expr


@dataclass
class Cmp(Statement):
    lhs: Expr
    rhs: Expr


@dataclass
class UnaryOp(Expr):
    op: Literal["-", "+"]
    operand: Expr


@dataclass
class Call(Expr):
    name: Literal["sin", "cos", "tan", "exp", "ln", "sqrt"]
    args: list[Expr]


@dataclass
class Constant(Expr):
    pass


@dataclass
class Number(Constant):
    value: Union[int, float]


@dataclass
class Pi(Constant):
    pass


@dataclass
class Name(Expr):
    id: str


@dataclass
class Bit(Expr):
    name: Name
    addr: int



from kirin import ir, types
from kirin.decl import info, statement

from .._dialect import dialect


@statement
class Measurement(ir.Statement):
    name = "measurement"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    """probability of noise introduced by measurement. For example 0.01 means 1% the measurement will be flipped"""
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


# 1Q measurements
@statement(dialect=dialect)
class MZ(Measurement):
    name = "MZ"


@statement(dialect=dialect)
class MY(Measurement):
    name = "MY"


@statement(dialect=dialect)
class MX(Measurement):
    name = "MX"


# Pair measurements
@statement(dialect=dialect)
class MZZ(Measurement):
    name = "MZZ"


@statement(dialect=dialect)
class MYY(Measurement):
    name = "MYY"


@statement(dialect=dialect)
class MXX(Measurement):
    name = "MXX"



from typing import Generic, TypeVar

from . import ast

T = TypeVar("T")

class Visitor(Generic[T]):
    def visit(self, node: ast.Node) -> T: ...
    def generic_visit(self, node: ast.Node) -> T: ...
    def visit_MainProgram(self, node: ast.MainProgram) -> T: ...
    def visit_OPENQASM(self, node: ast.OPENQASM) -> T: ...
    def visit_Kirin(self, node: ast.Kirin) -> T: ...
    def visit_QReg(self, node: ast.QReg) -> T: ...
    def visit_CReg(self, node: ast.CReg) -> T: ...
    def visit_Gate(self, node: ast.Gate) -> T: ...
    def visit_Opaque(self, node: ast.Opaque) -> T: ...
    def visit_IfStmt(self, node: ast.IfStmt) -> T: ...
    def visit_Cmp(self, node: ast.Cmp) -> T: ...
    def visit_Barrier(self, node: ast.Barrier) -> T: ...
    def visit_Include(self, node: ast.Include) -> T: ...
    def visit_Measure(self, node: ast.Measure) -> T: ...
    def visit_Reset(self, node: ast.Reset) -> T: ...
    def visit_Instruction(self, node: ast.Instruction) -> T: ...
    def visit_UGate(self, node: ast.UGate) -> T: ...
    def visit_CXGate(self, node: ast.CXGate) -> T: ...
    def visit_Bit(self, node: ast.Bit) -> T: ...
    def visit_BinOp(self, node: ast.BinOp) -> T: ...
    def visit_UnaryOp(self, node: ast.UnaryOp) -> T: ...
    def visit_Call(self, node: ast.Call) -> T: ...
    def visit_Number(self, node: ast.Number) -> T: ...
    def visit_Pi(self, node: ast.Pi) -> T: ...
    def visit_Name(self, node: ast.Name) -> T: ...
    # extensions
    def visit_ParaU3Gate(self, node: ast.ParaU3Gate) -> T: ...
    def visit_ParaCZGate(self, node: ast.ParaCZGate) -> T: ...
    def visit_ParaRZGate(self, node: ast.ParaRZGate) -> T: ...
    def visit_ParallelQArgs(self, node: ast.ParallelQArgs) -> T: ...
    def visit_GlobUGate(self, node: ast.GlobUGate) -> T: ...
    def visit_NoisePAULI1(self, node: ast.NoisePAULI1) -> T: ...



# Copyright (c) 2024, QuEra Computing Inc.
# All rights reserved.

from io import StringIO
from typing import Any, Dict, Optional
from dataclasses import dataclass

import pandas as pd
from pandas import DataFrame
from bloqade.visual.animation.runtime import qpustate as vis_qpustate

from .schema import NoiseModel


@dataclass
class QuEraSimulationResult:
    """Results of the QuEra hardware model simulation.

    Fields:
        flair_visual_version (str): The version of the Flair Visual package used to generate the simulation result.
        counts (dict[str, int]): The measurement bitstrings of the simulation.
        logs (DataFrame): Grainular logs events of what happened to each atom during the simulation.
        atom_animation_state (vis_qpustate.AnimateQPUState): Object used to play back atom trajectories and events during the simulation.
        noise_model (NoiseModel): The noise model used in the simulation.

    """

    flair_visual_version: str
    counts: dict[str, int]
    logs: DataFrame
    atom_animation_state: vis_qpustate.AnimateQPUState
    noise_model: NoiseModel

    @classmethod
    def from_json(cls, json: dict) -> "QuEraSimulationResult":
        """deserialize the object from a JSON serializable dictionary."""
        flair_visual_version = json["flair_visual_version"]
        counts = json["counts"]
        logs = pd.read_csv(StringIO(json["logs"]), index_col=0)
        atom_animation_state = vis_qpustate.AnimateQPUState.from_json(
            json["atom_animation_state"]
        )
        noise_model = NoiseModel(**json["noise_model"])

        return cls(
            flair_visual_version=flair_visual_version,
            counts=counts,
            logs=logs,
            atom_animation_state=atom_animation_state,
            noise_model=noise_model,
        )

    def to_json(self) -> Dict[str, Any]:
        """Turn the object into a JSON serializable dictionary."""
        return {
            "flair_visual_version": self.flair_visual_version,
            "counts": self.counts,
            "logs": self.logs.to_csv(),
            "atom_animation_state": self.atom_animation_state.to_json(),
            "noise_model": self.noise_model.model_dump(mode="json"),
        }

    def animate(
        self,
        dilation_rate: float = 0.05,
        fps: int = 30,
        gate_display_dilation: float = 1.0,
        save_mpeg: bool = False,
        filename: str = "vqpu_animation",
        start_block: int = 0,
        n_blocks: Optional[int] = None,
    ):
        """animate the qpu state

        Args:
            dilation_rate (float): Conversion factor from the qpu time to animation time units. when dilation_rate=1.0, 1 (us) of qpu exec time corresponds to 1 second of animation time.
            fps (int, optional): frame per second. Defaults to 30.
            gate_display_dilation (float, optional): relative dilation rate of a gate event. Defaults to 1. When setting higher value, the gate event will be displayed longer.
            save_mpeg (bool, optional): Save as mpeg. Defaults to False.
            filename (str, optional): The file name of saved mpeg file. Defaults to "vqpu_animation". When `save_mpeg` is False, this argument is ignored.
            start_block (int, optional): The start block to animate. Defaults to 0.
            n_blocks (int, optional): number of blocks to animate. Defaults to None. When None, animate all blocks after `start_block`.
        Returns:
            ani: matplotlib animation object
        """
        from bloqade.visual.animation.animate import animate_qpu_state

        ani = animate_qpu_state(
            state=self.atom_animation_state,
            dilation_rate=dilation_rate,
            fps=fps,
            gate_display_dilation=gate_display_dilation,
            start_block=start_block,
            n_blocks=n_blocks,
            save_mpeg=save_mpeg,
            filename=filename,
        )
        return ani



from kirin import ir, types
from kirin.decl import info, statement
from kirin.print import Printer

from .._dialect import dialect as dialect


@statement(dialect=dialect)
class ConstInt(ir.Statement):
    """IR Statement representing a constant integer value."""

    name = "constant.int"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), ir.FromPythonCall()})
    value: int = info.attribute(types.Int)
    """value (int): The constant integer value."""
    result: ir.ResultValue = info.result(types.Int)
    """result (Int): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class ConstFloat(ir.Statement):
    """IR Statement representing a constant float value."""

    name = "constant.float"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), ir.FromPythonCall()})
    value: float = info.attribute(types.Float)
    """value (float): The constant float value."""
    result: ir.ResultValue = info.result(types.Float)
    """result (Float): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class ConstBool(ir.Statement):
    """IR Statement representing a constant float value."""

    name = "constant.bool"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), ir.FromPythonCall()})
    value: bool = info.attribute(types.Bool)
    """value (float): The constant float value."""
    result: ir.ResultValue = info.result(types.Bool)
    """result (Float): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class ConstStr(ir.Statement):
    """IR Statement representing a constant str value."""

    name = "constant.str"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), ir.FromPythonCall()})
    value: str = info.attribute(types.String)
    """value (str): The constant str value."""
    result: ir.ResultValue = info.result(types.String)
    """result (str): The result value."""

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.plain_print(repr(self.value))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class Neg(ir.Statement):
    """IR Statement representing a negation operation."""

    name = "neg"
    traits = frozenset({ir.FromPythonCall()})
    operand: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(types.Int)



from kirin.emit import EmitStrFrame
from kirin.interp import MethodTable, impl
from bloqade.stim.emit.stim import EmitStimMain

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.stim")
class EmitStimNoiseMethods(MethodTable):

    single_p_error_map: dict[str, str] = {
        stmts.Depolarize1.name: "DEPOLARIZE1",
        stmts.Depolarize2.name: "DEPOLARIZE2",
        stmts.XError.name: "X_ERROR",
        stmts.YError.name: "Y_ERROR",
        stmts.ZError.name: "Z_ERROR",
    }

    @impl(stmts.XError)
    @impl(stmts.YError)
    @impl(stmts.ZError)
    @impl(stmts.Depolarize1)
    @impl(stmts.Depolarize2)
    def single_p_error(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.Depolarize1
    ):

        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        p: str = frame.get(stmt.p)
        name = self.single_p_error_map[stmt.name]
        res = f"{name}({p}) " + " ".join(targets)
        emit.writeln(frame, res)

        return ()

    @impl(stmts.PauliChannel1)
    def pauli_channel1(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.PauliChannel1
    ):

        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        px: str = frame.get(stmt.px)
        py: str = frame.get(stmt.py)
        pz: str = frame.get(stmt.pz)
        res = f"PAULI_CHANNEL_1({px},{py},{pz}) " + " ".join(targets)
        emit.writeln(frame, res)

        return ()

    @impl(stmts.PauliChannel2)
    def pauli_channel2(
        self, emit: EmitStimMain, frame: EmitStrFrame, stmt: stmts.PauliChannel2
    ):

        targets: tuple[str, ...] = frame.get_values(stmt.targets)
        prob: tuple[str, ...] = frame.get_values(stmt.args)[
            :15
        ]  # extract the first 15 argument, which is the probabilities
        prob_str: str = ", ".join(prob)

        res = f"PAULI_CHANNEL_2({prob_str}) " + " ".join(targets)
        emit.writeln(frame, res)

        return ()



from kirin import ir, types
from kirin.decl import info, statement

from ._dialect import dialect


@statement(dialect=dialect)
class Depolarize1(ir.Statement):
    name = "Depolarize1"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class Depolarize2(ir.Statement):
    name = "Depolarize2"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class PauliChannel1(ir.Statement):
    name = "PauliChannel1"
    traits = frozenset({ir.FromPythonCall()})
    px: ir.SSAValue = info.argument(types.Float)
    py: ir.SSAValue = info.argument(types.Float)
    pz: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class PauliChannel2(ir.Statement):
    name = "PauliChannel2"
    # TODO custom lowering to make sugar for this
    traits = frozenset({ir.FromPythonCall()})
    pix: ir.SSAValue = info.argument(types.Float)
    piy: ir.SSAValue = info.argument(types.Float)
    piz: ir.SSAValue = info.argument(types.Float)
    pxi: ir.SSAValue = info.argument(types.Float)
    pxx: ir.SSAValue = info.argument(types.Float)
    pxy: ir.SSAValue = info.argument(types.Float)
    pxz: ir.SSAValue = info.argument(types.Float)
    pyi: ir.SSAValue = info.argument(types.Float)
    pyx: ir.SSAValue = info.argument(types.Float)
    pyy: ir.SSAValue = info.argument(types.Float)
    pyz: ir.SSAValue = info.argument(types.Float)
    pzi: ir.SSAValue = info.argument(types.Float)
    pzx: ir.SSAValue = info.argument(types.Float)
    pzy: ir.SSAValue = info.argument(types.Float)
    pzz: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class XError(ir.Statement):
    name = "X_ERROR"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class YError(ir.Statement):
    name = "Y_ERROR"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)


@statement(dialect=dialect)
class ZError(ir.Statement):
    name = "Z_ERROR"
    traits = frozenset({ir.FromPythonCall()})
    p: ir.SSAValue = info.argument(types.Float)
    targets: tuple[ir.SSAValue, ...] = info.argument(types.Int)



from dataclasses import field, dataclass

from kirin import ir, interp
from bloqade.types import QubitType
from kirin.dialects import py, func, ilist
from kirin.ir.dialect import Dialect as Dialect
from bloqade.qasm2.parse import ast

from .base import EmitError, EmitQASM2Base, EmitQASM2Frame


def _default_dialect_group():
    from bloqade.qasm2.groups import gate

    return gate


@dataclass
class EmitQASM2Gate(EmitQASM2Base[ast.UOp | ast.Barrier, ast.Gate]):
    keys = ["emit.qasm2.gate"]
    dialects: ir.DialectGroup = field(default_factory=_default_dialect_group)


@ilist.dialect.register(key="emit.qasm2.gate")
class Ilist(interp.MethodTable):

    @interp.impl(ilist.New)
    def emit_ilist(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: ilist.New):
        return (ilist.IList(data=frame.get_values(stmt.values)),)


@py.constant.dialect.register(key="emit.qasm2.gate")
class Constant(interp.MethodTable):

    @interp.impl(py.Constant)
    def emit_constant(
        self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: py.Constant
    ):
        return (stmt.value,)


@func.dialect.register(key="emit.qasm2.gate")
class Func(interp.MethodTable):

    @interp.impl(func.Call)
    def emit_call(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: func.Call):
        raise EmitError("cannot emit dynamic call")

    @interp.impl(func.Invoke)
    def emit_invoke(
        self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: func.Invoke
    ):
        cparams, qparams = [], []
        for arg in stmt.inputs:
            if arg.type.is_subseteq(QubitType):
                qparams.append(frame.get(arg))
            else:
                cparams.append(frame.get(arg))

        frame.body.append(
            ast.Instruction(
                name=ast.Name(stmt.callee.sym_name),
                params=cparams,
                qargs=qparams,
            )
        )
        return ()

    @interp.impl(func.Lambda)
    @interp.impl(func.GetField)
    def emit_err(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt):
        raise EmitError(f"illegal statement {stmt.name} for QASM2 gate routine")

    @interp.impl(func.Return)
    @interp.impl(func.ConstantNone)
    def ignore(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt):
        return ()



from kirin import ir
from kirin.decl import info, statement
from bloqade.qasm2.types import QubitType
from bloqade.qasm2.dialects.expr.stmts import PyNum

from ._dialect import dialect


# trait
@statement
class SingleQubitGate(ir.Statement):
    """Base class for single qubit gates."""

    traits = frozenset({ir.FromPythonCall()})
    qarg: ir.SSAValue = info.argument(QubitType)
    """qarg (Qubit): The qubit argument."""


@statement
class TwoQubitCtrlGate(ir.Statement):
    traits = frozenset({ir.FromPythonCall()})
    ctrl: ir.SSAValue = info.argument(QubitType)
    """ctrl (Qubit): The control qubit."""
    qarg: ir.SSAValue = info.argument(QubitType)
    """qarg (Qubit): The target qubit."""


@statement(dialect=dialect)
class CX(TwoQubitCtrlGate):
    """Alias for the CNOT or CH gate operations."""

    name = "CX"  # Note this is capitalized


@statement(dialect=dialect)
class UGate(SingleQubitGate):
    """Apply A general single qubit unitary gate."""

    name = "U"
    theta: ir.SSAValue = info.argument(PyNum)
    """theta (float): The theta parameter."""
    phi: ir.SSAValue = info.argument(PyNum)
    """phi (float): The phi parameter."""
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The lambda parameter."""


@statement(dialect=dialect)
class Barrier(ir.Statement):
    """Apply the Barrier statement."""

    name = "barrier"
    traits = frozenset({ir.FromPythonCall()})
    qargs: tuple[ir.SSAValue, ...] = info.argument(QubitType)
    """qargs: tuple of qubits to apply the barrier to."""


# qelib1.inc as statements


@statement(dialect=dialect)
class Id(SingleQubitGate):
    """Apply the Identity gate."""

    name = "id"


@statement(dialect=dialect)
class H(SingleQubitGate):
    """Apply the Hadamard gate."""

    name = "h"


@statement(dialect=dialect)
class X(SingleQubitGate):
    """Apply the X gate."""

    name = "x"


@statement(dialect=dialect)
class Y(SingleQubitGate):
    """Apply the Y gate."""

    name = "y"


@statement(dialect=dialect)
class Z(SingleQubitGate):
    """Apply the Z gate."""

    name = "z"


@statement(dialect=dialect)
class S(SingleQubitGate):
    """Apply the S gate."""

    name = "s"


@statement(dialect=dialect)
class Sdag(SingleQubitGate):
    """Apply the hermitian conj of S gate."""

    name = "sdg"


@statement(dialect=dialect)
class SX(SingleQubitGate):
    """Apply the quantum Sqrt(X) gate."""

    name = "sx"


@statement(dialect=dialect)
class SXdag(SingleQubitGate):
    """Apply the dagger of quantum Sqrt(X) gate."""

    name = "sxdg"


@statement(dialect=dialect)
class T(SingleQubitGate):
    """Apply the T gate."""

    name = "t"


@statement(dialect=dialect)
class Tdag(SingleQubitGate):
    """Apply the hermitian conj of T gate."""

    name = "tdg"


@statement(dialect=dialect)
class RX(SingleQubitGate):
    """Apply the RX gate."""

    name = "rx"
    theta: ir.SSAValue = info.argument(PyNum)
    """theta (float): The angle of rotation around x axis."""


@statement(dialect=dialect)
class RY(SingleQubitGate):
    """Apply the RY gate."""

    name = "ry"
    theta: ir.SSAValue = info.argument(PyNum)
    """theta (float): The angle of rotation around y axis."""


@statement(dialect=dialect)
class RZ(SingleQubitGate):
    """Apply the RZ gate."""

    name = "rz"
    theta: ir.SSAValue = info.argument(PyNum)
    """theta (float): the angle of rotation around Z axis."""


@statement(dialect=dialect)
class U1(SingleQubitGate):
    """Apply the U1 gate."""

    name = "u1"
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The lambda parameter."""


@statement(dialect=dialect)
class U2(SingleQubitGate):
    """Apply the U2 gate."""

    name = "u2"
    phi: ir.SSAValue = info.argument(PyNum)
    """phi (float): The phi parameter."""
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The lambda parameter."""


@statement(dialect=dialect)
class CZ(TwoQubitCtrlGate):
    """Apply the Controlled-Z gate."""

    name = "cz"


@statement(dialect=dialect)
class CY(TwoQubitCtrlGate):
    """Apply the Controlled-Y gate."""

    name = "cy"


@statement(dialect=dialect)
class CSX(TwoQubitCtrlGate):
    """Apply the Controlled-Sqrt(X) gate."""

    name = "csx"


@statement(dialect=dialect)
class Swap(TwoQubitCtrlGate):
    """Apply the Swap gate."""

    name = "swap"


@statement(dialect=dialect)
class CH(TwoQubitCtrlGate):
    """Apply the Controlled-H gate."""

    name = "ch"


@statement(dialect=dialect)
class CCX(ir.Statement):
    """Apply the doubly controlled X gate."""

    name = "ccx"
    traits = frozenset({ir.FromPythonCall()})
    ctrl1: ir.SSAValue = info.argument(QubitType)
    """ctrl1 (Qubit): The first control qubit."""
    ctrl2: ir.SSAValue = info.argument(QubitType)
    """ctrl2 (Qubit): The second control qubit."""
    qarg: ir.SSAValue = info.argument(QubitType)
    """qarg (Qubit): The target qubit."""


@statement(dialect=dialect)
class CSwap(ir.Statement):
    """Apply the controlled swap gate."""

    name = "ccx"
    traits = frozenset({ir.FromPythonCall()})
    ctrl: ir.SSAValue = info.argument(QubitType)
    """ctrl (Qubit): The control qubit."""
    qarg1: ir.SSAValue = info.argument(QubitType)
    """qarg1 (Qubit): The first target qubit."""
    qarg2: ir.SSAValue = info.argument(QubitType)
    """qarg2 (Qubit): The second target qubit."""


@statement(dialect=dialect)
class CRX(TwoQubitCtrlGate):
    """Apply the Controlled-RX gate."""

    name = "crx"
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The angle to rotate around the X axis."""


@statement(dialect=dialect)
class CRY(TwoQubitCtrlGate):
    """Apply the Controlled-RY gate."""

    name = "cry"
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The angle to rotate around the Y axis."""


@statement(dialect=dialect)
class CRZ(TwoQubitCtrlGate):
    """Apply the Controlled-RZ gate."""

    name = "crz"
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The angle to rotate around the Z axis."""


@statement(dialect=dialect)
class CU1(TwoQubitCtrlGate):
    """Apply the Controlled-U1 gate."""

    name = "cu1"
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The lambda parameter."""


@statement(dialect=dialect)
class CU3(TwoQubitCtrlGate):
    """Apply the Controlled-U3 gate."""

    name = "cu3"
    theta: ir.SSAValue = info.argument(PyNum)
    phi: ir.SSAValue = info.argument(PyNum)
    """phi (float): The phi parameter."""
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The lambda parameter."""


@statement(dialect=dialect)
class CU(TwoQubitCtrlGate):
    """Apply the Controlled-U gate."""

    name = "cu"
    theta: ir.SSAValue = info.argument(PyNum)
    phi: ir.SSAValue = info.argument(PyNum)
    """phi (float): The phi parameter."""
    lam: ir.SSAValue = info.argument(PyNum)
    """lam (float): The lambda parameter."""
    gamma: ir.SSAValue = info.argument(PyNum)


@statement(dialect=dialect)
class RXX(TwoQubitCtrlGate):
    """Apply the XX rotation gate."""

    name = "rxx"
    theta: ir.SSAValue = info.argument(PyNum)
    """theta (float): The angle of rotation around the X axis."""


@statement(dialect=dialect)
class RZZ(TwoQubitCtrlGate):
    """Apply the ZZ rotation gate."""

    name = "rzz"
    theta: ir.SSAValue = info.argument(PyNum)
    """theta (float): The angle of rotation around the Z axis."""



from typing import Literal

from kirin import interp
from bloqade.qasm2.parse import ast
from bloqade.qasm2.types import QubitType
from kirin.emit.exceptions import EmitError
from bloqade.qasm2.emit.gate import EmitQASM2Gate, EmitQASM2Frame

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.qasm2.gate")
class EmitExpr(interp.MethodTable):

    @interp.impl(stmts.GateFunction)
    def emit_func(
        self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.GateFunction
    ):

        cparams, qparams = [], []
        for arg in stmt.body.blocks[0].args[1:]:
            name = frame.get(arg)
            if not isinstance(name, ast.Name):
                raise EmitError("expected ast.Name")
            if arg.type.is_subseteq(QubitType):
                qparams.append(name.id)
            else:
                cparams.append(name.id)
        emit.run_ssacfg_region(frame, stmt.body)
        emit.output = ast.Gate(
            name=stmt.sym_name,
            cparams=cparams,
            qparams=qparams,
            body=frame.body,
        )
        return ()

    @interp.impl(stmts.ConstInt)
    @interp.impl(stmts.ConstFloat)
    def emit_const_int(
        self,
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt: stmts.ConstInt | stmts.ConstFloat,
    ):
        return (ast.Number(stmt.value),)

    @interp.impl(stmts.ConstPI)
    def emit_const_pi(
        self,
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt: stmts.ConstPI,
    ):
        return (ast.Pi(),)

    @interp.impl(stmts.Neg)
    def emit_neg(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.Neg):
        arg = emit.assert_node(ast.Expr, frame.get(stmt.value))
        return (ast.UnaryOp("-", arg),)

    @interp.impl(stmts.Sin)
    @interp.impl(stmts.Cos)
    @interp.impl(stmts.Tan)
    @interp.impl(stmts.Exp)
    @interp.impl(stmts.Log)
    @interp.impl(stmts.Sqrt)
    def emit_sin(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt):
        arg = emit.assert_node(ast.Expr, frame.get(stmt.value))
        return (ast.Call(stmt.name, [arg]),)

    def emit_binop(
        self,
        sym: Literal["+", "-", "*", "/", "^"],
        emit: EmitQASM2Gate,
        frame: EmitQASM2Frame,
        stmt,
    ):
        lhs = emit.assert_node(ast.Expr, frame.get(stmt.lhs))
        rhs = emit.assert_node(ast.Expr, frame.get(stmt.rhs))
        return (ast.BinOp(sym, lhs, rhs),)

    @interp.impl(stmts.Add)
    def emit_add(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.Add):
        return self.emit_binop("+", emit, frame, stmt)

    @interp.impl(stmts.Sub)
    def emit_sub(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.Add):
        return self.emit_binop("-", emit, frame, stmt)

    @interp.impl(stmts.Mul)
    def emit_mul(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.Add):
        return self.emit_binop("*", emit, frame, stmt)

    @interp.impl(stmts.Div)
    def emit_div(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.Add):
        return self.emit_binop("/", emit, frame, stmt)

    @interp.impl(stmts.Pow)
    def emit_pow(self, emit: EmitQASM2Gate, frame: EmitQASM2Frame, stmt: stmts.Add):
        return self.emit_binop("^", emit, frame, stmt)



from kirin import interp
from kirin.analysis import ForwardFrame
from bloqade.analysis.schedule import DagScheduleAnalysis

from . import stmts
from ._dialect import dialect


@dialect.register(key="qasm2.schedule.dag")
class UOp(interp.MethodTable):

    @interp.impl(stmts.Id)
    @interp.impl(stmts.SXdag)
    @interp.impl(stmts.SX)
    @interp.impl(stmts.X)
    @interp.impl(stmts.Y)
    @interp.impl(stmts.Z)
    @interp.impl(stmts.H)
    @interp.impl(stmts.S)
    @interp.impl(stmts.Sdag)
    @interp.impl(stmts.T)
    @interp.impl(stmts.Tdag)
    @interp.impl(stmts.RX)
    @interp.impl(stmts.RY)
    @interp.impl(stmts.RZ)
    @interp.impl(stmts.U1)
    @interp.impl(stmts.U2)
    @interp.impl(stmts.UGate)
    def single_qubit_gate(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: stmts.SingleQubitGate,
    ):
        interp.update_dag(stmt, [stmt.qarg])
        return ()

    @interp.impl(stmts.RXX)
    @interp.impl(stmts.RZZ)
    @interp.impl(stmts.CX)
    @interp.impl(stmts.CY)
    @interp.impl(stmts.CZ)
    @interp.impl(stmts.CH)
    @interp.impl(stmts.CRZ)
    @interp.impl(stmts.CRY)
    @interp.impl(stmts.CRX)
    @interp.impl(stmts.CU1)
    @interp.impl(stmts.CU3)
    @interp.impl(stmts.CU)
    @interp.impl(stmts.CSX)
    def two_qubit_ctrl_gate(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: stmts.TwoQubitCtrlGate,
    ):
        interp.update_dag(stmt, [stmt.ctrl, stmt.qarg])
        return ()

    @interp.impl(stmts.CCX)
    def ccx_gate(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: stmts.CCX,
    ):
        interp.update_dag(stmt, [stmt.ctrl1, stmt.ctrl2, stmt.qarg])
        return ()

    @interp.impl(stmts.CSwap)
    def cswap_gate(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: stmts.CSwap,
    ):
        interp.update_dag(stmt, [stmt.ctrl, stmt.qarg1, stmt.qarg2])
        return ()

    @interp.impl(stmts.Barrier)
    def barrier(
        self,
        interp: DagScheduleAnalysis,
        frame: ForwardFrame,
        stmt: stmts.Barrier,
    ):
        interp.update_dag(stmt, stmt.qargs)
        return ()



from kirin import ir, passes
from kirin.prelude import structural_no_opt
from kirin.dialects import scf, func, ilist, lowering
from bloqade.qasm2.dialects import (
    uop,
    core,
    expr,
    glob,
    noise,
    inline,
    indexing,
    parallel,
)
from bloqade.qasm2.rewrite.desugar import IndexingDesugarPass


@ir.dialect_group([uop, func, expr, lowering.func, lowering.call])
def gate(self):
    fold_pass = passes.Fold(self)
    typeinfer_pass = passes.TypeInfer(self)

    def run_pass(
        method: ir.Method,
        *,
        fold: bool = True,
    ):
        method.verify()

        if isinstance(method.code, func.Function):
            new_code = expr.GateFunction(
                sym_name=method.code.sym_name,
                signature=method.code.signature,
                body=method.code.body,
            )
            method.code = new_code
        else:
            raise ValueError(
                "Gate Method code must be a Function, cannot be lambda/closure"
            )

        if fold:
            fold_pass(method)

        typeinfer_pass(method)
        method.code.typecheck()

    return run_pass


@ir.dialect_group(
    [
        uop,
        expr,
        core,
        scf,
        indexing,
        func,
        lowering.func,
        lowering.call,
    ]
)
def main(self):
    fold_pass = passes.Fold(self)
    typeinfer_pass = passes.TypeInfer(self)

    def run_pass(
        method: ir.Method,
        *,
        fold: bool = True,
    ):
        method.verify()
        # TODO make special Function rewrite

        if fold:
            fold_pass(method)

        typeinfer_pass(method)
        method.code.typecheck()

    return run_pass


@ir.dialect_group(
    structural_no_opt.union(
        [
            inline,
            uop,
            glob,
            noise,
            parallel,
            core,
        ]
    )
)
def extended(self):
    fold_pass = passes.Fold(self)
    typeinfer_pass = passes.TypeInfer(self)
    ilist_desugar_pass = ilist.IListDesugar(self)
    indexing_desugar_pass = IndexingDesugarPass(self)

    def run_pass(
        mt: ir.Method,
        *,
        fold: bool = True,
        typeinfer: bool = True,
    ):
        mt.verify()
        if fold:
            fold_pass.fixpoint(mt)

        if typeinfer:
            typeinfer_pass(mt)
        ilist_desugar_pass(mt)
        indexing_desugar_pass(mt)
        if typeinfer:
            typeinfer_pass(mt)  # fix types after desugaring
            mt.code.typecheck()

    return run_pass



from dataclasses import dataclass

from kirin import ir
from kirin.rewrite import abc, dce, walk, result, fixpoint
from kirin.passes.abc import Pass

from .stmts import PauliChannel, CZPauliChannel, AtomLossChannel
from ._dialect import dialect


class RemoveNoiseRewrite(abc.RewriteRule):
    def rewrite_Statement(self, node: ir.Statement) -> result.RewriteResult:
        if isinstance(node, (AtomLossChannel, PauliChannel, CZPauliChannel)):
            node.delete()
            return result.RewriteResult(has_done_something=True)

        return result.RewriteResult()


@dataclass
class RemoveNoisePass(Pass):
    name = "remove-noise"

    def unsafe_run(self, mt: ir.Method) -> result.RewriteResult:
        delete_walk = walk.Walk(RemoveNoiseRewrite())
        dce_walk = fixpoint.Fixpoint(walk.Walk(dce.DeadCodeElimination()))

        result = delete_walk.rewrite(mt.code)

        mt.dialects = ir.DialectGroup(mt.dialects.data.symmetric_difference([dialect]))

        if result.has_done_something:
            result = dce_walk.rewrite(mt.code)

        return result



from kirin import types, interp
from kirin.analysis import TypeInference
from kirin.dialects import py
from bloqade.qasm2.types import CRegType, QRegType, QubitType

from ._dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(py.indexing.GetItem, QRegType, types.Int)
    def getitem_qreg(
        self, infer: TypeInference, frame: interp.Frame, node: py.indexing.GetItem
    ):
        return (QubitType,)

    @interp.impl(py.indexing.GetItem, CRegType, types.Int)
    def getitem_creg(
        self, infer: TypeInference, frame: interp.Frame, node: py.indexing.GetItem
    ):
        return (QubitType,)



from bloqade import stim


def test_const():

    @stim.main
    def test_const_int():
        return 3

    out = test_const_int()

    assert out == 3
    assert isinstance(out, int)


def test_float():

    @stim.main
    def test_const_float():
        return 3.0

    out = test_const_float()

    assert out == 3.0
    assert isinstance(out, float)


def test_get_rec():

    @stim.main
    def get_rec():
        return stim.GetRecord(id=-3)

    get_rec.print()

    out = get_rec()

    assert isinstance(out, stim.RecordResult)
    assert out.value == -3



from bloqade import qasm2
from bloqade.qasm2 import glob
from bloqade.qasm2.passes import parallel
from bloqade.qasm2.rewrite import SimpleOptimalMergePolicy


def test_one():

    @qasm2.gate
    def gate(q1: qasm2.Qubit, q2: qasm2.Qubit):
        qasm2.cx(q1, q2)

    @qasm2.extended
    def test():
        q = qasm2.qreg(4)

        theta = 0.1
        phi = 0.2
        lam = 0.3

        qasm2.u(q[1], theta, phi, lam)
        qasm2.u(q[0], 0.4, phi, lam)
        qasm2.u(q[3], theta, phi, lam)
        qasm2.u(q[2], 0.4, phi, lam)

        gate(q[1], q[3])
        qasm2.barrier((q[1], q[2]))
        qasm2.u(q[2], theta, phi, lam)
        glob.u(theta=theta, phi=phi, lam=lam, registers=[q])
        qasm2.u(q[0], theta, phi, lam)

        gate(q[0], q[2])

    parallel.UOpToParallel(test.dialects)(test)
    test.print()


def test_two():

    @qasm2.extended
    def test():
        q = qasm2.qreg(8)

        qasm2.rz(q[0], 0.8)
        qasm2.rz(q[1], 0.7)
        qasm2.rz(q[2], 0.6)
        qasm2.rz(q[3], 0.5)
        qasm2.rz(q[4], 0.5)
        qasm2.rz(q[5], 0.6)
        qasm2.rz(q[6], 0.7)
        qasm2.rz(q[7], 0.8)

    parallel.UOpToParallel(test.dialects, SimpleOptimalMergePolicy)(test)
    test.print()


if __name__ == "__main__":
    test_one()
    test_two()
